# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


def rotation_matrix(alpha_deg, dim=3, axis=0):
    r"""Rotation matrix with given rotation axis and dimension (2d or 3d).

    Parameters
    ----------
    alpha_deg : int
        Rotation angle in degree.
    dim : int, optional (default is 3)
        Dimension of the rotation matrix.
    axis : int, optional (default is 0)
        Rotation axis.

    Returns
    -------
    rotation_matrix : ndarray
        Rotation matrix of dim 2 or 3 with given rotation axis.

    Notes
    -----
    The two-dimensional rotation axis is denoted in Eq. :eq:`rotation-matrix-2d`.

    ..  math::
        :label: rotation-matrix-2d

        \boldsymbol{R}(\alpha) = \begin{bmatrix}
            \cos(\alpha) & -\sin(\alpha) \\
            \sin(\alpha) &  \cos(\alpha)
        \end{bmatrix}

    A three-dimensional rotation matrix is created by inserting zeros in the row and
    column at the given axis of rotation and one at the intersection, see
    Eq. :eq:`rotation-matrix-3d`. If the axis of rotation is the second axis, the two-
    dimensinal rotation matrix is transposed.

    ..  math::
        :label: rotation-matrix-3d

        \boldsymbol{R}(\alpha) = \begin{bmatrix}
            \cos(\alpha) & -\sin(\alpha) & 0 \\
            \sin(\alpha) &  \cos(\alpha) & 0 \\
                  0      &        0      & 1
        \end{bmatrix}

    Examples
    --------
    >>> import numpy as np
    >>> import felupe as fem
    >>>
    >>> R = fem.math.rotation_matrix(alpha_deg=45, dim=2)
    >>> x = np.array([1., 0.])
    >>> y = R @ x
    >>> y
    array([0.70710678, 0.70710678])
    """

    a = np.deg2rad(alpha_deg)
    rotation_matrix = np.array([[np.cos(a), -np.sin(a)], [np.sin(a), np.cos(a)]])

    if dim == 3:
        if axis == 1:
            rotation_matrix = rotation_matrix.T
        rotation_matrix = np.insert(rotation_matrix, [axis], np.zeros((1, 2)), axis=0)
        rotation_matrix = np.insert(rotation_matrix, [axis], np.zeros((3, 1)), axis=1)
        rotation_matrix[axis, axis] = 1

    return rotation_matrix


def rotate_points(points, angle_deg, axis, center=None, mask=None):
    """Rotate points along a given axis.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    angle_deg : int
        Rotation angle in degree.
    axis : int
        Rotation axis.
    center : list or ndarray or None, optional
        Center point coordinates (default is None).
    mask : ndarray or None, optional
        A boolean mask to select points which are rotated (default is None).

    Returns
    -------
    points : ndarray
        Modified point coordinates.

    Examples
    --------
    Rotate the points of a rectangle in the xy-plane by 35 degree.

    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> points = fem.Rectangle(b=(3, 1), n=(10, 4)).points
       >>> points_new = fem.math.rotate_points(
       ...     points, angle_deg=35, axis=2, center=[1.5, 0.5]
       ... )

    See Also
    --------
    felupe.mesh.rotate : Rotate a Mesh.
    felupe.Mesh.rotate : Rotate a Mesh.
    """

    points = np.array(points)
    dim = points.shape[1]

    if center is None:
        center = np.zeros(dim)
    else:
        center = np.array(center)
    center = center.reshape(1, -1)

    if mask is None:
        mask = slice(None)

    points_rotated = (
        rotation_matrix(angle_deg, dim, axis) @ (points - center).T
    ).T + center

    points_new = points.copy()
    points_new[mask] = points_rotated[mask]

    return points_new


def revolve_points(points, n=11, phi=180, axis=0, expand_dim=True):
    """Revolve points along a given axis.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    n : int, optional
        Number of n-point revolutions (or (n-1) cell revolutions),
        default is 11.
    phi : float or ndarray, optional
        Revolution angle in degree (default is 180).
    axis : int, optional
        Revolution axis (default is 0).
    expand_dim : bool, optional
        Expand the dimension of the point coordinates (default is True).

    Returns
    -------
    points : ndarray
        Modified point coordinates.

    Examples
    --------
    Revolve the points of a cylinder from a rectangle.

    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> points = fem.Rectangle(a=(0, 4), b=(3, 5), n=(10, 4)).points
       >>> points_new = fem.math.revolve_points(mesh.points, n=11, phi=180, axis=0)

    See Also
    --------
    felupe.mesh.revolve : Revolve a 0d-Point to a 1d-Line, a 1d-Line to 2d-Quad or a
        2d-Quad to a 3d-Hexahedron Mesh.
    felupe.Mesh.revolve : Revolve a 2d-Quad to a 3d-Hexahedron Mesh.
    """

    points = np.array(points)
    dim = points.shape[1]

    if np.isscalar(phi):
        points_phi = np.linspace(0, phi, n)
    else:
        points_phi = phi
        n = len(points_phi)

    dim_new = dim
    if expand_dim:
        dim_new = dim + 1

    p = np.pad(points, ((0, 0), (0, dim_new - dim)))

    points_new = np.vstack(
        [(rotation_matrix(angle, dim_new, axis=axis) @ p.T).T for angle in points_phi]
    )

    if points_phi[-1] == 360:
        points_new = points_new[: len(points_new) - len(points)]

    return points_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import sqrt, trace
from tensortrax.math.linalg import det


def blatz_ko(C, mu):
    r"""Strain energy function of the Blatz-Ko isotropic hyperelastic
    `foam <https://doi.org/10.1122/1.548937>`_ material formulation [1]_.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    mu : float
        The shear modulus.

    Notes
    -----
    The Poisson ratio of the Blatz-Ko model formulation is :math:`\nu = 0.25`. The
    strain energy function is given in Eq. :eq:`psi-blatz-ko`

    ..  math::
        :label: psi-blatz-ko

        \psi = \frac{\mu}{2} \left(\frac{I_2}{I_3} + 2 \sqrt{I_3} - 5 \right)

    The shear modulus :math:`\mu` is related to young's modulus as denoted in Eq.
    :eq:`shear-modulus-blatz-ko`.

    ..  math::
        :label: shear-modulus-blatz-ko

        \mu = \frac{2 E}{5}

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(mat.models.hyperelastic.blatz_ko, mu=1.0)
        >>> ax = umat.plot()

    References
    ----------
    .. [1] P. J. Blatz and W. L. Ko, "Application of Finite Elastic Theory to the
       Deformation of Rubbery Materials", Transactions of the Society of Rheology, vol.
       6, no. 1. Society of Rheology, pp. 223–252, Mar. 01, 1962. doi: 10.1122/1.548937.
    """

    I1 = trace(C)
    I2 = (I1**2 - trace(C @ C)) / 2
    I3 = det(C)

    return mu * (I2 / I3 + 2 * sqrt(I3) - 5)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from tensortrax.math import abs as tensor_abs
from tensortrax.math import array, exp, maximum, sqrt
from tensortrax.math.special import try_stack


def morph_uniaxial(λ, statevars, p, ε=1e-6):
    """Return the force (per undeformed area) for a given longitudinal stretch in
    uniaxial incompressible tension or compression for the MORPH material
    formulation [1]_, [2]_.

    Parameters
    ----------
    λ : tensortrax.Tensor or jax.Array
        Longitudinal stretch of uniaxial incompressible deformation.
    statevars : array
        Vector of stacked state variables (CTS, λ - 1, SA1, SA2).
    p : list of float
        A list which contains the 8 material parameters.
    ε : float, optional
        A small stabilization parameter (default is 1e-6).

    References
    ----------
    .. [1] D. Besdo and J. Ihlemann, "A phenomenological constitutive model for
       rubberlike materials and its numerical applications", International Journal
       of Plasticity, vol. 19, no. 7. Elsevier BV, pp. 1019–1036, Jul. 2003. doi:
       `10.1016/s0749-6419(02)00091-8 <https://doi.org/10.1016/s0749-6419(02)00091-8>`_.

    .. [2] M. Freund, "Verallgemeinerung eindimensionaler Materialmodelle für die
       Finite-Elemente-Methode", Dissertation, Technische Universität Chemnitz,
       Chemnitz, 2013.

    """
    CTSn = array(statevars[:21], like=λ, shape=(21,))
    λn = array(statevars[21:42], like=λ, shape=(21,)) + 1
    SA1n = array(statevars[42:63], like=λ, shape=(21,))
    SA2n = array(statevars[63:84], like=λ, shape=(21,))

    CT = tensor_abs(λ**2 - 1 / λ)
    CTS = maximum(CT, CTSn)

    L1 = 2 * (λ**3 / λn - λn**2) / 3
    L2 = (λn**2 / λ**3 - 1 / λn) / 3
    LT = tensor_abs(L1 - L2)

    sigmoid = lambda x: 1 / sqrt(1 + x**2)
    α = p[0] + p[1] * sigmoid(p[2] * CTS)
    β = p[3] * sigmoid(p[2] * CTS)
    γ = p[4] * CTS * (1 - sigmoid(CTS / p[5]))

    L1_LT = L1 / (ε + LT)
    L2_LT = L2 / (ε + LT)
    CT_CTS = CT / (ε + CTS)

    SL1 = (γ * exp(p[6] * L1_LT * CT_CTS) + p[7] * L1_LT) / λ**2
    SL2 = (γ * exp(p[6] * L2_LT * CT_CTS) + p[7] * L2_LT) * λ

    SA1 = (SA1n + β * LT * SL1) / (1 + β * LT)
    SA2 = (SA2n + β * LT * SL2) / (1 + β * LT)

    dψdλ = (2 * α + SA1) * λ - (2 * α + SA2) / λ**2
    statevars_new = try_stack([CTS, (λ - 1), SA1, SA2], fallback=statevars)

    return dψdλ, statevars_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._base import ConstitutiveMaterial


class Material(ConstitutiveMaterial):
    r"""A user-defined material definition with given functions for the (first
    Piola-Kirchhoff) stress tensor :math:`\boldsymbol{P}`, optional constraints on
    additional fields (e.g. :math:`p` and :math:`J`), updated state variables
    :math:`\boldsymbol{\zeta}` as well as the according fourth-order elasticity tensor
    :math:`\mathbb{A}` and the linearizations of the constraint equations. Both
    functions take a list of the 3x3 deformation gradient :math:`\boldsymbol{F}` and
    optional vector of state variables :math:`\boldsymbol{\zeta}_n` as the first input
    argument. The stress-function must return the updated state variables
    :math:`\boldsymbol{\zeta}`.
    
    Parameters
    ----------
    stress : callable
        A constitutive material definition which returns a list containting the (first
        Piola-Kirchhoff) stress tensor, optional additional constraints as well as the
        state variables. The state variables must always be included even if they are
        None. See template code-blocks for the required function signature.
    elasticity : callable
        A constitutive material definition which returns a list containing the fourth-
        order elasticity tensor as the jacobian of the (first Piola-Kirchhoff) stress
        tensor w.r.t. the deformation gradient, optional linearizations of the
        additional constraints. The state variables must not be returned. See template
        code-blocks for the required function signature.
    nstatevars : int, optional
        Number of internal state variable components (default is 0). State variable
        components must always be concatenated into a 1d-array.

    Notes
    -----
    
    ..  note::
        The first item in the list of the input arguments always contains the
        gradient of the (displacement) field :math:`\boldsymbol{u}` w.r.t. the
        undeformed coordinates  :math:`\boldsymbol{X}`. The identity matrix
        :math:`\boldsymbol{1}` is added to this gradient, i.e. the first item of the
        list ``x`` contains the deformation gradient :math:`\boldsymbol{F} =
        \boldsymbol{1} + \frac{\partial \boldsymbol{u}}{\partial \boldsymbol{X}}`. All
        other fields are provided as interpolated values (no gradients evaluated).
    
    For :math:`(\boldsymbol{u})` single-field formulations, the callables for
    ``stress`` and ``elasticity`` must return the gradient and hessian of the strain
    energy density function :math:`\psi(\boldsymbol{F})` w.r.t. the deformation
    gradient tensor :math:`\boldsymbol{F}`.
    
    ..  math::

        \text{stress}(\boldsymbol{F}, \boldsymbol{\zeta}_n) = \begin{bmatrix}
            \frac{\partial \psi}{\partial \boldsymbol{F}} \\
            \boldsymbol{\zeta}
        \end{bmatrix}
    
    The callable for ``elasticity`` (hessian) must not return the updated state
    variables.
    
    ..  math::
        
        \text{elasticity}(\boldsymbol{F}, \boldsymbol{\zeta}_n) = \begin{bmatrix}
            \frac{\partial^2 \psi}{\partial \boldsymbol{F}\ \partial \boldsymbol{F}}
        \end{bmatrix}
    
    Take this code-block as template:

    ..  code-block::

        def stress(x, **kwargs):
            "First Piola-Kirchhoff stress tensor."

            # extract variables
            F, statevars = x[0], x[-1]

            # user code for (first Piola-Kirchhoff) stress tensor
            P = None

            # update state variables
            statevars_new = None

            return [P, statevars_new]

        def elasticity(x, **kwargs):
            "Fourth-order elasticity tensor."

            # extract variables
            F, statevars = x[0], x[-1]

            # user code for fourth-order elasticity tensor
            # according to the (first Piola-Kirchhoff) stress tensor
            dPdF = None

            return [dPdF]

        umat = Material(stress, elasticity, **kwargs)
        
    For :math:`(\boldsymbol{u}, p, J)` mixed-field formulations, the callables for
    ``stress`` and ``elasticity`` must return the gradients and hessians of the
    (augmented) strain energy density function w.r.t. the deformation gradient and the
    other fields.
    
    ..  math::

        \text{stress}(\boldsymbol{F}, p, J, \boldsymbol{\zeta}_n) = \begin{bmatrix}
            \frac{\partial \psi}{\partial \boldsymbol{F}} \\
            \frac{\partial \psi}{\partial p} \\
            \frac{\partial \psi}{\partial J} \\
            \boldsymbol{\zeta}
        \end{bmatrix}

    For the hessians, the upper-triangle blocks have to be provided.

    ..  math::

        \text{elasticity}(\boldsymbol{F}, p, J, \boldsymbol{\zeta}_n) = \begin{bmatrix}
            \frac{\partial^2 \psi}{\partial \boldsymbol{F}\ \partial \boldsymbol{F}} \\
            \frac{\partial^2 \psi}{\partial \boldsymbol{F}\ \partial p} \\
            \frac{\partial^2 \psi}{\partial \boldsymbol{F}\ \partial J} \\
            \frac{\partial^2 \psi}{\partial p\ \partial p} \\
            \frac{\partial^2 \psi}{\partial p\ \partial J} \\
            \frac{\partial^2 \psi}{\partial J\ \partial J}
        \end{bmatrix}

    For :math:`(\boldsymbol{u}, p, J)` mixed-field formulations, take this code-block as
    template:

    ..  code-block::

        def gradient(x, **kwargs):
            "Gradients of the strain energy density function."

            # extract variables
            F, p, J, statevars = x[0], x[1], x[2], x[-1]

            # user code
            dWdF = None  # first Piola-Kirchhoff stress tensor
            dWdp = None
            dWdJ = None

            # update state variables
            statevars_new = None

            return [dWdF, dWdp, dWdJ, statevars_new]

        def hessian(x, **kwargs):
            "Hessians of the strain energy density function."

            # extract variables
            F, p, J, statevars = x[0], x[1], x[2], x[-1]

            # user code
            d2WdFdF = None  # fourth-order elasticity tensor
            d2WdFdp = None
            d2WdFdJ = None
            d2Wdpdp = None
            d2WdpdJ = None
            d2WdJdJ = None

            # upper-triangle items of the hessian
            return [d2WdFdF, d2WdFdp, d2WdFdJ, d2Wdpdp, d2WdpdJ, d2WdJdJ]

        umat = Material(gradient, hessian, **kwargs)
    
    Examples
    --------
    The compressible isotropic hyperelastic Neo-Hookean material formulation is given
    by the strain energy density function
    
    .. math::

        \psi(\boldsymbol{C}) = \frac{\mu}{2} \text{tr}(\boldsymbol{C})
            - \mu \ln(J) + \frac{\lambda}{2} \ln(J)^2

    with the determinant of the deformation gradient and the right Cauchy Green
    deformation tensor.

    .. math::

       J &= \text{det}(\boldsymbol{F})
       
       C &= \boldsymbol{F}^T\ \boldsymbol{F}
      
    The first Piola-Kirchhoff stress tensor is evaluated as the gradient
    of the strain energy density function.

    .. math::

       \boldsymbol{P} &= \frac{\partial \psi}{\partial \boldsymbol{F}}

       \boldsymbol{P} &= \mu \left( \boldsymbol{F} - \boldsymbol{F}^{-T} \right)
           + \lambda \ln(J) \boldsymbol{F}^{-T}

    The hessian of the strain energy density function enables the corresponding
    elasticity tensor.

    .. math::

       \mathbb{A} &= \frac{\partial^2 \psi}{\partial \boldsymbol{F}\ \partial
       \boldsymbol{F}}

       \mathbb{A} &= \mu \boldsymbol{I} \overset{ik}{\otimes} \boldsymbol{I}
           + \left(\mu - \lambda \ln(J) \right)
               \boldsymbol{F}^{-T} \overset{il}{\otimes} \boldsymbol{F}^{-T}
           + \lambda \boldsymbol{F}^{-T} {\otimes} \boldsymbol{F}^{-T}

    ..  plot::
        :context: close-figs
    
        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> from felupe.math import (
        ...     cdya_ik,
        ...     cdya_il,
        ...     det,
        ...     dya,
        ...     identity,
        ...     inv,
        ...     transpose
        ... )
        >>>
        >>> def stress(x, mu, lmbda):
        ...     F, statevars = x[0], x[-1]
        ...     J = det(F)
        ...     lnJ = np.log(J)
        ...     iFT = transpose(inv(F, J))
        ...     dWdF = mu * (F - iFT) + lmbda * lnJ * iFT
        ...     return [dWdF, statevars]
        >>>
        >>> def elasticity(x, mu, lmbda):
        ...     F = x[0]
        ...     J = det(F)
        ...     iFT = transpose(inv(F, J))
        ...     eye = identity(F)
        ...     return [
        ...         mu * cdya_ik(eye, eye) + lmbda * dya(iFT, iFT) +
        ...         (mu - lmbda * np.log(J)) * cdya_il(iFT, iFT)
        ...     ]
        >>>
    
    The material formulation is tested in a minimal example of non-homogeneous uniaxial
    tension.
    
    ..  plot::
        :context: close-figs

        >>> mesh = fem.Cube(n=3)
        >>> region = fem.RegionHexahedron(mesh)
        >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
        >>>
        >>> umat = fem.Material(stress, elasticity, mu=1.0, lmbda=2.0)
        >>> solid = fem.SolidBodyNearlyIncompressible(umat, field, bulk=5000)
        >>>
        >>> boundaries = fem.dof.uniaxial(
        ...     field, clamped=True, move=0.5, return_loadcase=False
        ... )
        >>>
        >>> step = fem.Step(items=[solid], boundaries=boundaries)
        >>> job = fem.Job(steps=[step]).evaluate()
    
    See Also
    --------
    felupe.NeoHookeCompressible : Nearly-incompressible isotropic hyperelastic
        Neo-Hookean material formulation.

    """

    def __init__(self, stress, elasticity, nstatevars=0, **kwargs):
        self.umat = {"gradient": stress, "hessian": elasticity}
        self.kwargs = kwargs
        self.nstatevars = nstatevars
        self.x = [np.eye(3), np.zeros(nstatevars)]

    def gradient(self, x):
        """Return the evaluated gradient of the strain energy density function.

        Parameters
        ----------
        x : list of ndarray
            The list with input arguments. These contain the extracted fields of a
            :class:`~felupe.FieldContainer` along with the old vector of state
            variables, ``[*field.extract(), statevars_old]``.

        Returns
        -------
        list of ndarray
            A list with the evaluated gradient(s) of the strain energy density function
            and the updated vector of state variables.
        """

        return self.umat["gradient"](x, **self.kwargs)

    def hessian(self, x):
        """Return the evaluated upper-triangle components of the hessian(s) of the
        strain energy density function.

        Parameters
        ----------
        x : list of ndarray
            The list with input arguments. These contain the extracted fields of a
            :class:`~felupe.FieldContainer` along with the old vector of state
            variables, ``[*field.extract(), statevars_old]``.

        Returns
        -------
        list of ndarray
            A list with the evaluated upper-triangle components of the hessian(s) of the
            strain energy density function.
        """

        return self.umat["hessian"](x, **self.kwargs)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from tensortrax.math import trace
from tensortrax.math.linalg import det


def neo_hooke(C, mu):
    r"""Strain energy function of the isotropic hyperelastic
    `Neo-Hookean <https://en.wikipedia.org/wiki/Neo-Hookean_solid>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    mu : float
        Shear modulus.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-nh`.

    ..  math::
        :label: psi-nh

        \psi = \frac{\mu}{2} \left(\text{tr}\left(\hat{\boldsymbol{C}}\right) - 3\right)

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(mat.models.hyperelastic.neo_hooke, mu=1.0)
        >>> ax = umat.plot(incompressible=True)

    """
    return mu / 2 * (det(C) ** (-1 / 3) * trace(C) - 3)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import array, diag, sqrt
from jax.numpy import sum as asum
from jax.numpy.linalg import eigvalsh

from ....tensortrax.models.hyperelastic import storakers as storakers_docstring


@wraps(storakers_docstring)
def storakers(C, mu, alpha, beta):
    λ1, λ2, λ3 = sqrt(eigvalsh(C + diag(array([0, -1e-4, 1e-4]))))
    J = λ1 * λ2 * λ3

    μ = array(mu)
    α = array(alpha)
    β = array(beta)

    return asum(2 * μ / α**2 * (λ1**α + λ2**α + λ3**α - 3 + (J ** (-α * β) - 1) / β))



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import inspect
from functools import wraps

import jax
import jax.numpy as jnp
from jax.numpy.linalg import det


def vmap(fun, in_axes=0, out_axes=0, method=jax.vmap, **kwargs):
    """Vectorizing map. Creates a function which maps ``fun`` over argument axes. This
    decorator treats all non-specified arguments and keyword-arguments as static.

    See Also
    --------
    jax.vmap : Vectorizing map. Creates a function which maps ``fun`` over argument
        axes.
    """

    @wraps(fun)
    def vmap_with_static_kwargs(*args, **keywordargs):
        # sorted list of all parameter keys, including kwargs with default values
        sig = inspect.signature(fun)
        keys = [
            key
            for key, value in sig.parameters.items()
            if not (key in ["args", "kwargs"] and value.default == inspect._empty)
        ]

        if not (
            "kwargs" in sig.parameters.keys()
            and sig.parameters["kwargs"].default == inspect._empty
        ):
            # check if unexpected keyword-argument is given
            for key in keywordargs.keys():
                if key not in keys:
                    raise TypeError(
                        f"{fun.__name__}() got an unexpected keyword argument '{key}'"
                    )

        # dict with default values for all parameters
        parameters = dict(
            [(key, value.default) for key, value in sig.parameters.items()]
        )

        # merge dict of default values with custom keyword arguments
        items = {**parameters, **keywordargs}

        # create sorted list of values of keyword-arguments, including default kwargs
        keyword_args = [items[key] for key in keys[len(args) :]]

        # don't map non-given arguments and keyword-arguments
        if not hasattr(in_axes, "__len__"):
            in_axes_tuple = (in_axes,)
        else:
            in_axes_tuple = in_axes

        static_argnums = len(args) + len(keyword_args) - len(in_axes_tuple)
        in_axes_new = (*in_axes_tuple, *([None] * static_argnums))

        vfun = method(fun, in_axes=in_axes_new, out_axes=out_axes, **kwargs)

        return vfun(*args, *keyword_args)

    return vmap_with_static_kwargs


def vmap2(fun, in_axes=[0, 0], out_axes=[0, 0], methods=[jax.vmap, jax.vmap], **kwargs):
    "Nested vectorizing map."
    return vmap(
        vmap(
            fun, in_axes=in_axes[0], out_axes=out_axes[0], method=methods[0], **kwargs
        ),
        in_axes=in_axes[1],
        out_axes=out_axes[1],
        method=methods[1],
        **kwargs,
    )


def as_total_lagrange(fun):
    @wraps(fun)
    def evaluate(F, *args, **kwargs):
        i, j = jnp.triu_indices(3)
        C_triu = jnp.einsum("ia,ia->a", F[:, i], F[:, j])
        C = C_triu[jnp.array([[0, 1, 2], [1, 3, 4], [2, 4, 5]])]
        return fun(C, *args, **kwargs)

    return evaluate


def isochoric_volumetric_split(fun):
    """Apply the material formulation only on the isochoric part of the
    multiplicative split of the deformation gradient."""

    @wraps(fun)
    def apply_iso(C, *args, **kwargs):
        return fun(det(C) ** (-1 / 3) * C, *args, **kwargs)

    return apply_iso



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import log
from tensortrax.math import sum as tsum
from tensortrax.math import trace
from tensortrax.math.linalg import det, eigvalsh


def extended_tube(C, Gc, delta, Ge, beta):
    r"""Strain energy function of the isotropic hyperelastic
    `Extended Tube <https://www.doi.org/10.5254/1.3538822>`_ [1]_ material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    Gc : float
        Cross-link contribution to the initial shear modulus.
    delta : float
         Finite extension parameter of the polymer strands.
    Ge : float
        Constraint contribution to the initial shear modulus.
    beta : float
        Global rearrangements of cross-links upon deformation (release of topological
        constraints).

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-et`

    ..  math::
        :label: psi-et

        \psi = \frac{G_c}{2} \left[ \frac{\left( 1 - \delta^2 \right)
            \left( \hat{I}_1 - 3 \right)}{1 - \delta^2 \left( \hat{I}_1 - 3 \right)} +
            \ln \left( 1 - \delta^2 \left( \hat{I}_1 - 3 \right) \right) \right] +
            \frac{2 G_e}{\beta^2} \left( \hat{\lambda}_1^{-\beta} +
            \hat{\lambda}_2^{-\beta} + \hat{\lambda}_3^{-\beta} - 3 \right)

    with the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor as given in Eq. :eq:`invariants-et`

    ..  math::
        :label: invariants-et

        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

    and the principal stretches, obtained from the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`stretches-et`.

    ..  math::
        :label: stretches-et

        \lambda^2_\alpha &= \text{eigvals}\left( \boldsymbol{C} \right)

        \hat{\lambda}_\alpha &= J^{-1/3} \lambda_\alpha

    The initial shear modulus results from the sum of the cross-link and the constraint
    contributions to the total initial shear modulus as denoted in Eq.
    :eq:`shear-modulus-et`.

    ..  math::
        :label: shear-modulus-et

        \mu = G_e + G_c

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.extended_tube,
        ...     Gc=0.1867,
        ...     Ge=0.2169,
        ...     beta=0.2,
        ...     delta=0.09693,
        ... )
        >>> ax = umat.plot(incompressible=True)

    References
    ----------
    ..  [1] M. Kaliske and G. Heinrich, "An Extended Tube-Model for Rubber Elasticity:
        Statistical-Mechanical Theory and Finite Element Implementation", Rubber
        Chemistry and Technology, vol. 72, no. 4. Rubber Division, ACS, pp. 602–632,
        Sep. 01, 1999. doi:
        `10.5254/1.3538822 <https://www.doi.org/10.5254/1.3538822>`_.

    """
    J3 = det(C) ** (-1 / 3)
    D = J3 * trace(C)
    wC = J3 * eigvalsh(C)
    g = (1 - delta**2) * (D - 3) / (1 - delta**2 * (D - 3))
    Wc = Gc / 2 * (g + log(1 - delta**2 * (D - 3)))
    We = 2 * Ge / beta**2 * tsum(wC ** (-beta / 2) - 1)
    return Wc + We



"""
Strain-energy density functions for strain energy-gradient (stress) model formulations.

This module contains material model formulations to be used as the ``fun``-argument in
:func:`~felupe.constitution.jax.Material`. The gradient as well as the hessian of
the strain energy density function is carried out by automatic differentiation using
:mod:`jax`. Hence, all math-functions must be taken from :mod:`jax.numpy`.
"""

from ._morph import morph
from ._morph_representative_directions import morph_representative_directions
from ._morph_uniaxial import morph_uniaxial

__all__ = [
    "morph",
    "morph_representative_directions",
    "morph_uniaxial",
]

# default (stable) material parameters
morph.kwargs = dict(p=[0, 0, 0, 0, 0, 1, 0, 0])
morph_representative_directions.kwargs = dict(p=[0, 0, 0, 0, 0, 1, 0, 0])
morph_uniaxial.kwargs = dict(p=[0, 0, 0, 0, 0, 1, 0, 0])



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import warnings

import jax
import numpy as np

from .._material import Material as MaterialDefault
from ._helpers import vmap2


class Material(MaterialDefault):
    r"""A material definition with a given function for the partial derivative of the
    strain energy function w.r.t. the deformation gradient tensor with Automatic
    Differentiation provided by :mod:`jax`.

    Parameters
    ----------
    fun : callable
        A gradient of the strain energy density function w.r.t. the deformation gradient
        tensor :math:`\boldsymbol{F}`. Function signature must be
        ``fun = lambda F, **kwargs: P`` for functions without state variables and
        ``fun = lambda F, statevars, **kwargs: [P, statevars_new]`` for functions
        with state variables. It is important to use only differentiable math-functions
        from :mod:`jax`.
    nstatevars : int, optional
        Number of state variables (default is 0).
    jit : bool, optional
        A flag to invoke just-in-time compilation (default is True).
    parallel : bool, optional
        A flag to invoke parallel function evaluations (default is False). If True, the
        quadrature points are executed in parallel. The number of devices must be
        greater or equal the number of quadrature points per cell.
    jacobian : callable or None, optional
        A callable for the Jacobian. Default is None, where :func:`jax.jacobian` is
        used. This may be used to switch to forward-mode differentian
        :func:`jax.jacfwd`.
    **kwargs : dict, optional
        Optional keyword-arguments for the gradient of the strain energy density
        function.

    Notes
    -----
    The gradient of the strain energy density function
    :math:`\frac{\partial \psi}{\partial \boldsymbol{F}}` must be given in terms of the
    deformation gradient tensor :math:`\boldsymbol{F}`.

    ..  warning::
        It is important to use only differentiable math-functions from :mod:`jax`!

    Take this code-block as template

    ..  code-block::

        import felupe as fem
        import felupe.constitution.jax as mat
        import jax.numpy as jnp

        def neo_hooke(F, mu):
            "First Piola-Kirchhoff stress of the Neo-Hookean material formulation."

            C = F.T @ F
            Cu = jnp.linalg.det(C) ** (-1/3) * C
            dev = lambda C: C - jnp.trace(C) / 3 * jnp.eye(3)

            return mu * F @ dev(Cu) @ jnp.linalg.inv(C)

        umat = mat.Material(neo_hooke, mu=1)

    and this code-block for material formulations with state variables:

    ..  code-block::

        import felupe as fem
        import felupe.constitution.jax as mat
        import jax.numpy as jnp

        def viscoelastic(F, Cin, mu, eta, dtime):
            "Finite strain viscoelastic material formulation."

            # unimodular part of the right Cauchy-Green deformation tensor
            C = F.T @ F
            Cu = jnp.linalg.det(C) ** (-1 / 3) * C

            # update of state variables by evolution equation
            from_triu = lambda C: C[jnp.array([[0, 1, 2], [1, 3, 4], [2, 4, 5]])]
            Ci = from_triu(Cin) + mu / eta * dtime * Cu
            Ci = jnp.linalg.det(Ci) ** (-1 / 3) * Ci

            # second Piola-Kirchhoff stress tensor
            dev = lambda C: C - jnp.trace(C) / 3 * jnp.eye(3)
            S = mu * dev(Cu @ jnp.linalg.inv(Ci)) @ jnp.linalg.inv(C)

            # first Piola-Kirchhoff stress tensor and state variable
            i, j = jnp.triu_indices(3)
            to_triu = lambda C: C[i, j]
            return F @ S, to_triu(Ci)

        umat = mat.Material(viscoelastic, mu=1, eta=1, dtime=1, nstatevars=6)

    ..  note::
        See the `documentation of JAX <https://jax.readthedocs.io>`_ for further
        details. JAX uses single-precision (32bit) data types by default. This requires
        to relax the tolerance of :func:`~felupe.newtonrhapson` to ``tol=1e-4``. If
        required, JAX may be enforced to use double-precision at startup with
        ``jax.config.update("jax_enable_x64", True)``.

    Examples
    --------
    View force-stretch curves on elementary incompressible deformations.

    ..  plot::

        >>> import felupe as fem
        >>> import felupe.constitution.jax as mat
        >>> import jax.numpy as jnp
        >>>
        >>> def neo_hooke(F, mu):
        ...     "First Piola-Kirchhoff stress of the Neo-Hookean material formulation."
        ...
        ...     C = F.T @ F
        ...     Cu = jnp.linalg.det(C) ** (-1/3) * C
        ...     dev = lambda C: C - jnp.trace(C) / 3 * jnp.eye(3)
        ...
        ...     return mu * F @ dev(Cu) @ jnp.linalg.inv(C)
        >>>
        >>> umat = mat.Material(neo_hooke, mu=1)
        >>> ax = umat.plot(incompressible=True)

    """

    def __init__(
        self, fun, nstatevars=0, jit=True, parallel=False, jacobian=None, **kwargs
    ):
        has_aux = nstatevars > 0
        self.fun = fun

        if jacobian is None:
            jacobian = jax.jacobian

        keyword_args = kwargs
        if hasattr(fun, "kwargs"):
            keyword_args = {**fun.kwargs, **keyword_args}

        super().__init__(
            stress=self._stress,
            elasticity=self._elasticity,
            nstatevars=nstatevars,
            **keyword_args,
        )

        in_axes = out_axes_grad = [2, 3]
        if nstatevars > 0:
            in_axes = out_axes_grad = [(2, 1), (3, 2)]

        out_axes_hess = [4, 5]
        if nstatevars > 0:
            out_axes_hess = [(4, 1), (5, 2)]

        methods = [jax.vmap, jax.vmap]
        if parallel:
            methods[0] = jax.pmap  # apply on quadrature-points
            jit = False  # pmap uses jit

        self._grad = vmap2(
            self.fun, in_axes=in_axes, out_axes=out_axes_grad, methods=methods
        )
        self._hess = vmap2(
            jacobian(self.fun, has_aux=has_aux),
            in_axes=in_axes,
            out_axes=out_axes_hess,
            methods=methods,
        )

        if jit:
            self._grad = jax.jit(self._grad)
            self._hess = jax.jit(self._hess)

    def _stress(self, x, **kwargs):
        if self.nstatevars > 0:
            statevars = x[1]

        F = x[0]
        if self.nstatevars > 0:
            dWdF, statevars_new = self._grad(F, statevars, **kwargs)
            statevars_new = np.array(statevars_new)
        else:
            dWdF = self._grad(F, **kwargs)
            statevars_new = None

        return [np.array(dWdF), statevars_new]

    def _elasticity(self, x, **kwargs):
        if self.nstatevars > 0:
            statevars = x[1]

        F = x[0]
        if self.nstatevars > 0:
            d2WdFdF, statevars_new = self._hess(F, statevars, **kwargs)
        else:
            d2WdFdF = self._hess(F, **kwargs)
        return [np.array(d2WdFdF)]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from tensortrax import Tensor


def total_lagrange(material):
    r"""Decorate a second Piola-Kirchhoff stress Total-Lagrange material formulation as
    a first Piola-Kirchoff stress function.

    Notes
    -----
    ..  math::

        \delta \psi = \boldsymbol{F} \boldsymbol{S} : \delta \boldsymbol{F}

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>> import felupe.constitution.tensortrax as mat
        >>> import tensortrax.math as tm
        >>>
        >>> @fem.total_lagrange
        ... def neo_hooke_total_lagrange(F, mu=1):
        ...     C = F.T @ F
        ...     S = mu * tm.special.dev(tm.linalg.det(C)**(-1/3) * C) @ tm.linalg.inv(C)
        ...     return S
        >>>
        >>> umat = mat.Material(neo_hooke_total_lagrange, mu=1)

    See Also
    --------
    felupe.constitution.tensortrax.Hyperelastic : A hyperelastic material definition
        with a given function for the strain energy density function per unit undeformed
        volume with Automatic Differentiation.
    felupe.constitution.tensortrax.Material : A material definition with a given
        function for the partial derivative of the strain energy function w.r.t. the
        deformation gradient tensor with Automatic Differentiation.
    """

    @wraps(material)
    def first_piola_kirchhoff_stress(F, *args, **kwargs):
        # evaluate the second Piola-Kirchhoff stress
        res = material(F, *args, **kwargs)

        # check if the material formulation returns state variables and extract
        # the second Piola-Kirchhoff stress tensor
        if isinstance(res, Tensor):
            S = res
            statevars_new = None
        else:
            S, statevars_new = res

        # first Piola-Kirchhoff stress tensor
        P = F @ S

        if statevars_new is None:
            return P
        else:
            return P, statevars_new

    return first_piola_kirchhoff_stress



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ....math import cdya_ik, dev, dya, identity, trace


def linear_elastic_viscoelastic(dε, εn, σn, ζn, λ, μ, G, τ, Δt, **kwargs):
    r"""3D linear-elastic viscoelastic material formulation to be used in
    :class:`~felupe.MaterialStrain`.

    Arguments
    ---------
    dε : ndarray
        Strain increment.
    εn : ndarray
        Old strain tensor.
    σn : ndarray
        Old stress tensor.
    ζn : list
        List of old state variables.
    λ : float
        First Lamé-constant.
    μ : float
        Second Lamé-constant (shear modulus).
    G : list of float
        List of deviatoric viscoelastic shear moduli.
    τ : list of float
        List of time constants for deviatoric viscoelastic shear moduli.
    Δt : float
        Time increment.
    **kwargs
        Additional keyword arguments.

        tangent : bool
            A flag to evaluate the elasticity tensor.

    Returns
    -------
    dσdε : ndarray
        Elasticity tensor.
    σ : ndarray
        (New) stress tensor.
    ζ : list
        List of new state variables.

    Notes
    -----
    The stress consists of a long-term elastic and a deviatoric viscoelastic stress
    part, see Eq. :eq:`total-stress`.

    ..  math::
        :label: total-stress

        \boldsymbol{\sigma} = \boldsymbol{\sigma}_e + \boldsymbol{\sigma}_v

    The long-term elastic part is given in Eq. :eq:`elastic-stress`.

    ..  math::
        :label: elastic-stress

        \boldsymbol{\sigma}_e = 2 \mu\ \boldsymbol{\varepsilon}
            + \lambda \operatorname{tr}(\boldsymbol{\varepsilon}) \boldsymbol{1}

    The i-th viscous part is given in Eq. :eq:`visco-stress`,

    ..  math::
        :label: visco-stress

        \dot{\boldsymbol{\sigma}}_{v,i}
            + \frac{1}{\tau_i} \boldsymbol{\sigma}_{v,i} &=
            2 G_i \operatorname{dev}\left( \dot{\boldsymbol{\varepsilon}} \right)

        \boldsymbol{\sigma}_{v,i} &= a_i\ \boldsymbol{\sigma}_{v,i}^n + b_i
            \operatorname{dev} \left(
                \boldsymbol{\varepsilon} - \boldsymbol{\varepsilon}_n
            \right)

    with the coefficients as denoted in Eq. :eq:`visco-coeff`.

    ..  math::
        :label: visco-coeff

        a_i &= \exp \left( -\Delta t / \tau_i \right)

        b_i &= \frac{2 G_i\ \tau_i (1 - a_i)}{\Delta t}

    The total fourth-order elasticity tensor is given in Eq. :eq:`fourth-order-visco`.

    ..  math::
        :label: fourth-order-visco

        \mathbb{C} = 2 \bar{\mu} \ \boldsymbol{1} \odot \boldsymbol{1} +
                \bar{\lambda}\ \boldsymbol{1} \otimes \boldsymbol{1}

    The effective coefficients are summarized in Eq. :eq:`visco-coeff-tangent`.

    ..  math::
        :label: visco-coeff-tangent

        \bar{\mu} &= \mu + \sum_{i=1}^N \frac{b_i}{2}

        \bar{\lambda} &= \lambda - \frac{2}{3} \left( \bar{\mu} - \mu \right)

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.MaterialStrain(
        ...     material=fem.linear_elastic_viscoelastic,
        ...     λ=2.0,
        ...     μ=1.0,
        ...     G=[3.0, 2.0],
        ...     τ=[10.0, 100.0],
        ...     Δt=5.0,
        ...     statevars=((2, 3, 3),),
        ... )
        >>> ux = fem.math.linsteps([1, 2, 1, 2, 1, 2, 1], num=20)
        >>> ax = umat.plot(ux=ux, bx=None, ps=None)

    See Also
    --------
    MaterialStrain : A strain-based user-defined material definition with a given
        function for the stress tensor and the (fourth-order) elasticity tensor.

    """

    # helpers
    eye = identity(dim=3, shape=(1, 1))

    # new elastic stress (old stress σn is not used because of included viscoelasticity)
    ε = εn + dε
    σ = 2 * μ * ε + λ * trace(ε) * eye

    # update deviatoric viscoelastic stress
    dev_dε = dev(dε)

    a = [np.exp(-Δt / τi) for τi in τ]
    b = [2 * Gi * (1 - ai) * τi / Δt for Gi, ai, τi in zip(G, a, τ)]

    # Alternative method: backward-Euler
    # a = [1 / (1 + Δt / τi) for τi in τ]
    # b = [2 * Gi * ai for Gi, ai in zip(G, a)]

    # update of new deviatoric viscoelastic stress
    σvn = ζn[0]
    σv = σvn.copy()
    for i, σvni in enumerate(ζn[0]):
        σv[i] = a[i] * σvni + b[i] * dev_dε

    # evaluate elasticity tensor
    if kwargs["tangent"]:
        μ_eff = μ + np.sum(b, axis=0) / 2
        λ_eff = λ - 2 / 3 * (μ_eff - μ)
        dσdε = 2 * μ_eff * cdya_ik(eye, eye) + λ_eff * dya(eye, eye)
    else:
        dσdε = None

    # new total stress
    σ += np.sum(σv, axis=0)

    # new state variables
    ζ = [σv]

    return dσdε, σ, ζ



"""
constitution.hyperelasticity
============================
This module contains manually (no automatic differentiation) defined isotropic
hyperelastic constitutive material formulations with gradients and hessians of strain
energy density functions. For more sophisticated material model formulations see
:class:`~felupe.Hyperelastic`.
"""

from ._neo_hooke_compressible import NeoHookeCompressible
from ._neo_hooke_nearly_incompressible import NeoHooke
from ._ogden_roxburgh import OgdenRoxburgh
from ._volumetric import Volumetric

__all__ = [
    "NeoHooke",
    "NeoHookeCompressible",
    "OgdenRoxburgh",
    "Volumetric",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""


import inspect

import numpy as np

from ..assembly import IntegralForm
from ..math import dot, dya, identity, sqrt
from ._helpers import Assemble, Evaluate, Results
from ._solidbody import Solid


class TrussBody(Solid):
    r"""A Truss body with methods for the assembly of sparse vectors/matrices.

    Parameters
    ----------
    umat : class
        A class which provides methods for evaluating the gradient and the hessian of
        the strain energy density function per unit undeformed volume. The function
        signatures must be ``dψdλ, ζ_new = umat.gradient([λ, ζ])`` for
        the gradient and ``d2ψdλdλ = umat.hessian([λ, ζ])`` for the hessian of
        the strain energy density function :math:`\psi(\lambda)`, where
        :math:`\lambda` is the stretch and :math:`\zeta`
        holds the array of internal state variables.
    field : FieldContainer
        A field container with one or more fields.
    area : float or ndarray
        The cross-sectional areas of the trusses.
    statevars : ndarray or None, optional
        Array of initial internal state variables (default is None).

    Notes
    -----
    For a truss element the stretch is evaluated as given in Eq. :eq:`truss-stretch`

    .. math::
       :label: truss-stretch

       \Lambda = \sqrt{\frac{l^2}{L^2}}

    with the squared undeformed and deformed lengths, denoted in
    Eqs. :eq:`truss-lengths`.

    .. math::
       :label: truss-lengths

       l^2 &= \boldsymbol{\Delta x}^T \boldsymbol{\Delta x} \\
       L^2 &= \boldsymbol{\Delta X}^T \boldsymbol{\Delta X}

    This enables the Biot strain measure, see Eq. :eq:`truss-strain`.

    .. math::
       :label: truss-strain

       E_{11} = \Lambda - 1

    The normal force of a truss depends directly on the geometric exactly defined
    strain measure :math:`E_{11}`. For the general case of a nonlinear material
    behaviour the normal force is defined as given in Eq. :eq:`truss-force`

    .. math::
       :label: truss-force

       N = S_{11}(E_{11})~A

    and the derivative according to Eq. :eq:`truss-force-derivative`.

    .. math::
       :label: truss-force-derivative

       \frac{\partial N}{\partial E_{11}} = \frac{
               \partial S_{11}(E_{11})
           }{\partial E_{11}}~A

    For the case of a linear elastic material this reduces to
    Eqs. :eq:`truss-force-linear`.

    .. math::
       :label: truss-force-linear

       S_{11}(E_{11}) &= E~E_{11} \\
       N              &= EA~E_{11} \\
       \frac{\partial N}{\partial E_{11}} &= EA

    The (nonlinear) fixing force column vector may be expressed as a function of the
    elemental force :math:`N_k` and the deformed unit vector :math:`\boldsymbol{n}_k`,
    see Eq. :eq:`truss-fixing-force`.

    .. math::
       :label: truss-fixing-force

       \boldsymbol{r}_k = \begin{bmatrix}
               \boldsymbol{r}_A \\
               \boldsymbol{r}_E
           \end{bmatrix} = N_k \begin{pmatrix}
               -\boldsymbol{n}_k \\
               \phantom{-}\boldsymbol{n}_k
           \end{pmatrix}

    For a truss the stiffness matrix is divided into four block matrices of the
    same components but with different signs, see Eq. :eq:`truss-stiffness-matrix`.

    .. math::
       :label: truss-stiffness-matrix

       \boldsymbol{K}_{k~(6,6)} = \begin{bmatrix}
               \boldsymbol{K}_{AA} & \boldsymbol{K}_{AE}\\
               \boldsymbol{K}_{EA} & \boldsymbol{K}_{EE}
           \end{bmatrix} = \begin{pmatrix}
               \phantom{-}\boldsymbol{K}_{EE} & -\boldsymbol{K}_{EE}\\
               -\boldsymbol{K}_{EE} &  \phantom{-}\boldsymbol{K}_{EE}
           \end{pmatrix}

    A change in the fixing force vector at the end node `E` w.r.t. a small change of
    the displacements at node `E` is defined as the tangent stiffnes `EE`, see
    Eq. :eq:`truss-stiffness-block`.

    .. math::
       :label: truss-stiffness-block

       \boldsymbol{K}_{EE} &= \frac{
           \partial \boldsymbol{r}_E}{\partial \boldsymbol{U}_E
       } \\
       \boldsymbol{K}_{EE} &= \frac{A}{L} ~ \frac{
           \partial S_{11}(E_{11})
       }{\partial E_{11}} \boldsymbol{n} \otimes \boldsymbol{n} + \frac{N}{l} \left(
           \boldsymbol{1} - \boldsymbol{n} \otimes \boldsymbol{n}
       \right)

    Examples
    --------
    ..  pyvista-plot::
        :context:
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Mesh(
        ...     points=[[0, 0], [1, 1], [2.0, 0]],
        ...     cells=[[0, 1], [1, 2]],
        ...     cell_type="line",
        ... )
        >>> region = fem.RegionTruss(mesh)
        >>> field = fem.Field(region, dim=2).as_container()
        >>> boundaries = fem.BoundaryDict(fixed=fem.Boundary(field[0], fy=0))
        >>>
        >>> umat = fem.LinearElastic1D(E=[1, 1])
        >>> truss = fem.TrussBody(umat, field, area=[1, 1])
        >>> load = fem.PointLoad(field, points=[1])
        >>>
        >>> table = fem.math.linsteps([0, 1], num=5, axis=1, axes=2)
        >>> ramp = {load: table * -0.1}
        >>> step = fem.Step(items=[truss, load], ramp=ramp, boundaries=boundaries)
        >>> job = fem.Job(steps=[step]).evaluate()
        >>>
        >>> mesh.plot(
        ...     plotter=load.plot(plotter=boundaries.plot(), scale=0.5),
        ...     line_width=5
        ... ).show()

    """

    def __init__(self, umat, field, area, statevars=None):
        self.field = field
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)
        self.evaluate = Evaluate(gradient=self._gradient, hessian=self._hessian)
        self.results = Results(stress=True, elasticity=True)

        self.umat = umat
        self.area = np.array(area)

        self.form_vector = IntegralForm(
            [None],
            v=self.field,
            dV=None,
            grad_v=[False],
        )
        self.form_matrix = IntegralForm(
            [None],
            v=self.field,
            dV=None,
            u=self.field,
            grad_v=[False],
            grad_u=[False],
        )

        if statevars is not None:
            self.results.statevars = statevars
        else:
            statevars_shape = (0,)
            if hasattr(umat, "x"):
                statevars_shape = umat.x[-1].shape
            self.results.statevars = np.zeros(
                (
                    *statevars_shape,
                    field.region.mesh.ncells,
                )
            )

        cells = self.field.region.mesh.cells
        self.X = self.field.region.mesh.points[cells].T
        dX = self.X[:, 1] - self.X[:, 0]
        self.length_undeformed = sqrt(dot(dX, dX, mode=(1, 1)))

    def _kinematics(self, field):

        cells = self.field.region.mesh.cells
        u = field[0].values[cells].T
        x = self.X + u

        dx = x[:, 1] - x[:, 0]

        length_deformed = sqrt(dot(dx, dx, mode=(1, 1)))

        stretch = length_deformed / self.length_undeformed
        normal_deformed = dx / length_deformed

        return stretch, length_deformed, normal_deformed

    def _gradient(self, field=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field
            self.results.kinematics = self._kinematics(field)

        if "out" in inspect.signature(self.umat.gradient).parameters:
            kwargs["out"] = self.results.gradient

        gradient = self.umat.gradient(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )
        self.results.gradient = self.results.stress = gradient[0]
        self.results._statevars = gradient[-1]

        return self.results.gradient

    def _hessian(self, field=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field
            self.results.kinematics = self._kinematics(field)

        if "out" in inspect.signature(self.umat.hessian).parameters:
            kwargs["out"] = self.results.hessian

        hessian = self.umat.hessian(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )
        self.results.hessian = self.results.elasticity = hessian[0]

        return self.results.hessian

    def _vector(self, field=None, parallel=None, args=(), kwargs=None):

        self.results.stress = self._gradient(field, args=args, kwargs=kwargs)
        normal_deformed = self.results.kinematics[2]

        force = self.results.stress * self.area
        r = force * np.array([-normal_deformed, normal_deformed])  # (a, i, cell)

        self.results.force = self.form_vector.assemble(values=[r], parallel=parallel)

        return self.results.force

    def _matrix(self, field=None, parallel=None, args=(), kwargs=None):

        self.results.hessian = self._hessian(field, args=args, kwargs=kwargs)
        L = self.length_undeformed
        stretch, l, n = self.results.kinematics

        S = self.results.stress = self._gradient(field, args=args, kwargs=kwargs)
        dSdE = self.results.elasticity = self._hessian(field, args=args, kwargs=kwargs)

        m = dya(n, n, mode=1)
        eye = identity(dim=len(n), shape=(1,))
        K_EE = dSdE / L * self.area * m + S / l * self.area * (eye - m)

        K = np.stack([[K_EE, -K_EE], [-K_EE, K_EE]])  # (a, b, i, j, cell)

        self.results.stiffness = self.form_matrix.assemble(
            values=[K.transpose([0, 2, 1, 3, 4])], parallel=parallel
        )

        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from numpy.linalg import inv


def lame_converter(E, nu):
    r"""Convert the pair of given material parameters Young's modulus :math:`E` and
    Poisson ratio :math:`\nu` to first and second Lamé - constants :math:`\lambda` and
    :math:`\mu`.

    Parameters
    ----------
    E : float
        Young's modulus.
    nu : float
        Poisson ratio.

    Returns
    -------
    lmbda : float
        First Lamé - constant.
    mu : float
        Second Lamé - constant (shear modulus).

    Notes
    -----

    ..  math::

        \lambda &= \frac{E \nu}{(1 + \nu) (1 - 2 \nu)}

        \mu &= \frac{E}{2 (1 + \nu)}
    """

    lmbda = E * nu / ((1 + nu) * (1 - 2 * nu))
    mu = E / (2 * (1 + nu))

    return lmbda, mu


def lame_converter_orthotropic(E, nu, G):
    r"""Convert elastic orthotropic material parameters to Lamé constants.

    Parameters
    ----------
    E : list of float
        List of the three elastic moduli :math:`E_1, E_2, E_3`.
    nu : list of float
        List of three poisson ratios :math:`\nu_{12}, \nu_{23}, \nu_{31}`.
    G : list of float
        List of three shear moduli :math:`G_{12}, G_{23}, G_{31}`.

    Returns
    -------
    lmbda : list of float
        List of six (upper triangle) first Lamé parameters :math:`\lambda_{11},
        \lambda_{12}, \lambda_{13}, \lambda_{22}, \lambda_{23}, \lambda_{33}`.
    mu : list of float
        List of the three second Lamé parameters :math:`\mu_1,\mu_2, \mu_3`.

    Notes
    -----
    The orthotropic material parameters are converted to orthotropic Lamé
    constants.

    The compliance matrix as the inverse of the stiffness matrix with the
    parameters :math:`E_i`, :math:`\nu_{ij}` and :math:`G_{ij}` is given in
    Eq. :eq:`ortho-matrix-inv`.

    ..  math::
        :label: ortho-matrix-inv

        \boldsymbol{C}^{-1} = \begin{bmatrix}
            \frac{1}{E_1} & -\frac{\nu_{21}}{E_2} & -\frac{\nu_{31}}{E_3}
                & 0 & 0 & 0 \\
            -\frac{\nu_{12}}{E_1} & \frac{1}{E_2} & -\frac{\nu_{32}}{E_3}
                & 0 & 0 & 0 \\
            -\frac{\nu_{13}}{E_1} & -\frac{\nu_{23}}{E_2} & \frac{1}{E_3}
                & 0 & 0 & 0 \\
            0 & 0 & 0 & \frac{1}{G_{12}} & 0 & 0 \\
            0 & 0 & 0 & 0 & \frac{1}{G_{23}} & 0 \\  
            0 & 0 & 0 & 0 & 0 & \frac{1}{G_{31}}
        \end{bmatrix}

    The stiffness matrix with the Lamé constants is denoted in
    Eq. :eq:`ortho-matrix`.

    ..  math::
        :label: ortho-matrix-inv

        \boldsymbol{C} = \begin{bmatrix}
            \lambda_{11} + 2 \mu_1 & \lambda_{12} & \lambda_{13} & 0 & 0 & 0 \\
            \lambda_{11} & \lambda_{12} + 2 \mu_2 & \lambda_{13} & 0 & 0 & 0 \\
            \lambda_{11} & \lambda_{12} & \lambda_{13} + 2 \mu_3 & 0 & 0 & 0 \\
            0 & 0 & 0 & \frac{\mu_1 + \mu_2}{2} & 0 & 0 \\
            0 & 0 & 0 & 0 & \frac{\mu_2 + \mu_3}{2} & 0 \\  
            0 & 0 & 0 & 0 & 0 & \frac{\mu_3 + \mu_1}{2}
        \end{bmatrix}

    Eq. :eq:`ortho-matrix-inv` is evaluated and inverted numerically to extract
    the Lamé constants.

    See Also
    --------
    felupe.LinearElasticOrthotropic : Orthotropic linear-elastic
        material formulation.
    felupe.constitution.tensortrax.models.hyperelastic.saint_venant_kirchhoff_orthotropic :
        Strain energy function of the orthotropic hyperelastic Saint-Venant
        Kirchhoff material formulation.
    """

    # unpack orthotropic elastic material parameters
    E1, E2, E3 = E
    ν12, ν23, ν31 = nu
    G12, G23, G31 = G

    # orthotropic symmetry
    ν21 = ν12 * E2 / E1
    ν32 = ν23 * E3 / E2
    ν13 = ν31 * E1 / E3

    C = inv(
        [
            [1 / E1, -ν21 / E2, -ν31 / E3, 0, 0, 0],
            [-ν12 / E1, 1 / E2, -ν32 / E3, 0, 0, 0],
            [-ν13 / E1, -ν23 / E2, 1 / E3, 0, 0, 0],
            [0, 0, 0, 1 / G12, 0, 0],
            [0, 0, 0, 0, 1 / G23, 0],
            [0, 0, 0, 0, 0, 1 / G31],
        ]
    )

    # take the components from this matrix
    #
    # [λ11 + 2 * μ1, λ12, λ13, 0, 0, 0]
    # [λ12, λ22 + 2 * μ2, λ23, 0, 0, 0]
    # [λ13, λ23, λ33 + 2 * μ3, 0, 0, 0]
    # [0, 0, 0, (μ1 + μ2) / 2, 0, 0]
    # [0, 0, 0, 0, (μ2 + μ3) / 2, 0]
    # [0, 0, 0, 0, 0, (μ1 + μ3) / 2]

    λ12 = C[0, 1]
    λ23 = C[1, 2]
    λ13 = C[0, 2]

    μ1 = C[3, 3] - C[4, 4] + C[5, 5]
    μ2 = C[3, 3] + C[4, 4] - C[5, 5]
    μ3 = -C[3, 3] + C[4, 4] + C[5, 5]

    λ11 = C[0, 0] - 2 * μ1
    λ22 = C[1, 1] - 2 * μ2
    λ33 = C[2, 2] - 2 * μ3

    lmbda = [λ11, λ12, λ13, λ22, λ23, λ33]
    mu = [μ1, μ2, μ3]

    return lmbda, mu



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse.linalg import spsolve

from ..math import values


def partition(v, K, dof1, dof0, r=None):
    """Perform partitioning of field values (unknowns), (stiffness) matrix
    and (residuals) vector with given lists of active (dof1) and
    prescribed degrees of freedom (dof0).

    Parameters
    ----------
    v : FieldContainer
        A field container with the fields. It is used to extract and concatenate the
        1d-array of unknows.
    K : scipy.sparse.spmatrix
        A two-dimensional sparse (stiffness) matrix.
    dof1 : list of int
        List of active degrees of freedom (zero-indexed).
    dof0 : list of int
        List of prescribed degrees of freedom (zero-indexed).
    r : 1d-array or None, optional
        The 1d-vector of residuals. If None, the residuals are set to zero. Default is
        None.

    Returns
    -------
    u : 1d-array
        The concatenated full 1d-array of unknowns, extracted from the field container.
    u0 : 1d-array
        The prescribed unknowns.
    K11 : scipy.sparse.spmatrix
        The active-active block of the two-dimensional sparse (stiffness) matrix.
    K10 : scipy.sparse.spmatrix
        The active-prescribed block of the two-dimensional sparse (stiffness) matrix.
    dof1 : list of int
        List of active degrees of freedom (zero-indexed).
    dof0 : list of int
        List of prescribed degrees of freedom (zero-indexed).
    r1 : 1d-array
        The 1d-vector of active residuals.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=3)
        >>> region = fem.RegionQuad(mesh)
        >>> field = fem.FieldPlaneStrain(region, dim=2).as_container()
        >>> boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
        >>> umat = fem.NeoHooke(mu=1.0, bulk=2.0)
        >>> solid = fem.SolidBody(umat, field)
        >>>
        >>> K = solid.assemble.matrix()
        >>> r = solid.assemble.vector()
        >>>
        >>> dof0 = loadcase["dof0"]
        >>> dof1 = loadcase["dof1"]
        >>> ext0 = loadcase["ext0"]
        >>>
        >>> system = fem.solve.partition(field, K, dof1, dof0, r)

    See Also
    --------
    felupe.solve.solve : Linear solution of equation system.
    """

    # extract values
    u = values(v)

    # check if residuals vector is passed
    if r is None:
        r1 = None
    else:
        # partition residuals vector
        r1 = r[dof1]

    # prescribed dofs of unknowns
    u0 = u.ravel()[dof0]

    # partition (stiffness) matrix
    K11 = K[dof1, :][:, dof1]
    K10 = K[dof1, :][:, dof0]

    return u, u0, K11, K10, dof1, dof0, r1


def solve(u, u0, K11, K10, dof1, dof0, r1=None, ext0=None, solver=spsolve):
    r"""Linear solution of equation system with optional given values of
    unknowns at prescribed deegrees of freedom.

    Notes
    -----

    ..  math::

        \boldsymbol{K}_{11}\ d\boldsymbol{u}_1 =
            -\boldsymbol{r}_1 - \boldsymbol{K}_{10} (
                \boldsymbol{u}_{0,\text{ext}} - \boldsymbol{u}_0
            )

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=3)
        >>> region = fem.RegionQuad(mesh)
        >>> field = fem.FieldPlaneStrain(region, dim=2).as_container()
        >>> boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
        >>> umat = fem.NeoHooke(mu=1.0, bulk=2.0)
        >>> solid = fem.SolidBody(umat, field)
        >>>
        >>> K = solid.assemble.matrix()
        >>> r = solid.assemble.vector()
        >>>
        >>> dof0 = loadcase["dof0"]
        >>> dof1 = loadcase["dof1"]
        >>> ext0 = loadcase["ext0"]
        >>>
        >>> system = fem.solve.partition(field, K, dof1, dof0, r)
        >>> field += fem.solve.solve(*system, ext0=ext0)

    See Also
    --------
    felupe.solve.partition : Perform a partitioning into active and prescribed dof.
    """

    # init active residuals
    if r1 is None:
        r1 = np.zeros(len(dof1))

    # init external displacements
    if ext0 is None:
        ext0 = 0
        # init inactive dofs of residuals
        dr0 = np.zeros(len(dof1))
    else:
        # evaluate inactive dofs of residuals
        dr0 = K10.dot(ext0 - u0)

    # solve linear system (active dofs)
    du1 = solver(K11, -r1 - dr0.reshape(*r1.shape))

    # full solution
    du = np.empty(u.size)
    du[dof1] = du1
    du[dof0] = ext0 - u0

    # reshape solution to shape of input
    return du.reshape(*u.shape)



from ._field import ViewField
from ._mesh import ViewMesh
from ._scene import Scene
from ._solid import ViewSolid
from ._xdmf import ViewXdmf

__all__ = [
    "Scene",
    "ViewField",
    "ViewMesh",
    "ViewSolid",
    "ViewXdmf",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy as copy

import numpy as np

from ..math import dot, eigh, transpose
from ._view import ViewMaterial, ViewMaterialIncompressible


class ConstitutiveMaterial:
    r"""Base class for constitutive materials.

    A constitutive material definition, or so-called ``umat`` (user material), is a
    class with methods for evaluating gradients and hessians of the strain energy
    density function with respect to the defined fields in the field container, where
    the gradient of the first (displacement) field is passed as the deformation
    gradient. For all following fields, the field values (no gradients) are provided. An
    attribute ``x=[np.zeros(statevars_shape)]`` has to be added to the class to define
    the shape of optional state variables. For reasons of performance, FElupe passes the
    field gradients and values *all at once*, e.g. the deformation gradient is of shape
    ``(3, 3, q, c)``, where ``q`` refers to the number of quadrature points per cell and
    ``c`` to the number of cells. These last two axes are the so-called *trailing axes*.
    Math-functions from :ref:`felupe.math <felupe-api-math>` all support the operation
    on trailing axes. The constitutive material definition class should be inherited
    from :class:`~felupe.ConstitutiveMaterial` in order to provide force-stretch curves
    for elementary deformations.

    Examples
    --------
    Take this code-block as a template for a two-field :math:`(\boldsymbol{u}, p)`
    formulation with the old displacement gradient as a state variable:

    ..  pyvista-plot::
        :context:

        import numpy as np
        import felupe as fem

        # math-functions which support trailing axes
        from felupe.math import det, dya, identity, transpose, inv

        class MyMaterialFormulation(fem.ConstitutiveMaterial):

            def __init__(self):
                # provide the shape of state variables without trailing axes
                # values are ignored - state variables are always initiated with zeros
                self.x = [np.zeros((3, 3))]

            def gradient(self, x):
                "Gradients of the strain energy density function."

                # extract variables
                F, p, statevars = x[0], x[1], x[-1]

                # user code
                dWdF = None  # first Piola-Kirchhoff stress tensor
                dWdp = None

                # update state variables
                # example: the displacement gradient
                statevars_new = F - identity(F)

                return [dWdF, dWdp, statevars_new]

            def hessian(self, x, **kwargs):
                "Hessians of the strain energy density function."

                # extract variables
                F, p, statevars = x[0], x[1], x[-1]

                # user code
                d2WdFdF = None  # fourth-order elasticity tensor
                d2WdFdp = None
                d2Wdpdp = None

                # upper-triangle items of the hessian
                return [d2WdFdF, d2WdFdp, d2Wdpdp]

        umat = MyMaterialFormulation()

    See Also
    --------
    felupe.constitutive_material : A decorator for a constitutive material definition.
    """

    def copy(self):
        "Return a deep-copy of the constitutive material."
        return copy(self)

    def view(self, incompressible=False, **kwargs):
        """Create views on normal force per undeformed area vs. stretch curves for the
        elementary homogeneous deformations uniaxial tension/compression, planar shear
        and biaxial tension of a given isotropic material formulation.

        Parameters
        ----------
        incompressible : bool, optional
            A flag to enforce views on incompressible deformations (default is False).
        **kwargs : dict, optional
            Optional keyword-arguments for :class:`~felupe.ViewMaterial` or
            :class:`~felupe.ViewMaterialIncompressible`.

        Returns
        -------
        felupe.ViewMaterial or felupe.ViewMaterialIncompressible

        See Also
        --------
        felupe.ViewMaterial : Create views on normal force per undeformed area vs.
            stretch curves for the elementary homogeneous deformations uniaxial
            tension/compression, planar shear and biaxial tension of a given isotropic
            material formulation.
        felupe.ViewMaterialIncompressible : Create views on normal force per undeformed
            area vs. stretch curves for the elementary homogeneous incompressible
            deformations uniaxial tension/compression, planar shear and biaxial tension
            of a given isotropic material formulation.
        """

        View = ViewMaterial
        if incompressible:
            View = ViewMaterialIncompressible

        return View(self, **kwargs)

    def plot(self, incompressible=False, **kwargs):
        """Return a plot with normal force per undeformed area vs. stretch curves for
        the elementary homogeneous deformations uniaxial tension/compression, planar
        shear and biaxial tension of a given isotropic material formulation.

        Parameters
        ----------
        incompressible : bool, optional
            A flag to enforce views on incompressible deformations (default is False).
        **kwargs : dict, optional
            Optional keyword-arguments for :class:`~felupe.ViewMaterial` or
            :class:`~felupe.ViewMaterialIncompressible`.

        Returns
        -------
        matplotlib.axes.Axes

        See Also
        --------
        felupe.ViewMaterial : Create views on normal force per undeformed area vs.
            stretch curves for the elementary homogeneous deformations uniaxial
            tension/compression, planar shear and biaxial tension of a given isotropic
            material formulation.
        felupe.ViewMaterialIncompressible : Create views on normal force per undeformed
            area vs. stretch curves for the elementary homogeneous incompressible
            deformations uniaxial tension/compression, planar shear and biaxial tension
            of a given isotropic material formulation.
        """

        return self.view(incompressible=incompressible, **kwargs).plot()

    def screenshot(self, filename="umat.png", incompressible=False, **kwargs):
        """Save a screenshot with normal force per undeformed area vs. stretch curves
        for the elementary homogeneous deformations uniaxial tension/compression, planar
        shear and biaxial tension of a given isotropic material formulation.

        Parameters
        ----------
        filename : str, optional
            The filename of the screenshot (default is "umat.png").
        incompressible : bool, optional
            A flag to enforce views on incompressible deformations (default is False).
        **kwargs : dict, optional
            Optional keyword-arguments for :class:`~felupe.ViewMaterial` or
            :class:`~felupe.ViewMaterialIncompressible`.

        Returns
        -------
        matplotlib.axes.Axes

        See Also
        --------
        felupe.ViewMaterial : Create views on normal force per undeformed area vs.
            stretch curves for the elementary homogeneous deformations uniaxial
            tension/compression, planar shear and biaxial tension of a given isotropic
            material formulation.
        felupe.ViewMaterialIncompressible : Create views on normal force per undeformed
            area vs. stretch curves for the elementary homogeneous incompressible
            deformations uniaxial tension/compression, planar shear and biaxial tension
            of a given isotropic material formulation.
        """

        import matplotlib.pyplot as plt

        ax = self.plot(incompressible=incompressible, **kwargs)
        fig = ax.get_figure()
        fig.savefig(filename)
        plt.close(fig)

        return ax

    def optimize(
        self, ux=None, ps=None, bx=None, incompressible=False, relative=False, **kwargs
    ):
        r"""Optimize the material parameters by a least-squares fit on experimental
        stretch-stress data.

        Parameters
        ----------
        ux : array of shape (2, ...) or None, optional
            Experimental uniaxial stretch and force-per-undeformed-area data (default is
            None).
        ps : array of shape (2, ...) or None, optional
            Experimental planar-shear stretch and force-per-undeformed-area data
            (default is None).
        bx : array of shape (2, ...) or None, optional
            Experimental biaxial stretch and force-per-undeformed-area data (default is
            None).
        incompressible : bool, optional
            A flag to enforce incompressible deformations (default is False).
        relative : bool, optional
            A flag to optimize relative instead of absolute residuals, i.e.
            ``(predicted - observed) / observed`` instead of ``predicted - observed``
            (default is False).
        **kwargs : dict, optional
            Optional keyword arguments are passed to
            :func:`scipy.optimize.least_squares`.

        Returns
        -------
        ConstitutiveMaterial
            A copy of the constitutive material with the optimized material parameters.
        scipy.optimize.OptimizeResult
            Represents the optimization result.


        Notes
        -----
        ..  warning::
            At least one load case, i.e. one of the arguments ``ux``, ``ps`` or ``bx``
            must not be ``None``.

        The vector of residuals is given in Eq. :eq:`material-optimize-residuals` in
        case of absolute residuals

        ..  math::
            :label: material-optimize-residuals

            \boldsymbol{r} &= \begin{bmatrix}
                \boldsymbol{r}_\text{ux} \\
                \boldsymbol{r}_\text{ps} \\
                \boldsymbol{r}_\text{bx}
            \end{bmatrix}

            r_\text{ux}(\lambda_i) &= P_\text{ux}(\lambda_i)
                - P_\text{ux, observed}(\lambda_i)

            r_\text{ps}(\lambda_i) &= P_\text{ps}(\lambda_i)
                - P_\text{ps, observed}(\lambda_i)

            r_\text{bx}(\lambda_i) &= P_\text{bx}(\lambda_i)
                - P_\text{bx, observed}(\lambda_i)

        and in Eq. :eq:`material-optimize-residuals-relative` in case of relative
        residuals.

        ..  math::
            :label: material-optimize-residuals-relative

            \boldsymbol{r} &= \begin{bmatrix}
                \boldsymbol{r}_\text{ux} \\
                \boldsymbol{r}_\text{ps} \\
                \boldsymbol{r}_\text{bx}
            \end{bmatrix}

            r_\text{ux}(\lambda_i)  &= \frac{
                P_\text{ux}(\lambda_i) - P_\text{ux, observed}(\lambda_i)}{
                    P_\text{ux, observed}(\lambda_i)
                }

            r_\text{ps}(\lambda_i)  &= \frac{
                P_\text{ps}(\lambda_i) - P_\text{ps, observed}(\lambda_i)}{
                    P_\text{ps, observed}(\lambda_i)
                }

            r_\text{bx}(\lambda_i)  &= \frac{
                P_\text{bx}(\lambda_i) - P_\text{bx, observed}(\lambda_i)}{
                    P_\text{bx, observed}(\lambda_i)
                }

        Examples
        --------
        The :func:`Anssari-Benam Bucchi <felupe.anssari_benam_bucchi>` material model
        formulation is best-fitted on Treloar's uniaxial and biaxial tension data [1]_.

        ..  plot::

            >>> import numpy as np
            >>> import felupe as fem
            >>>
            >>> λ, P = np.array(
            ...     [
            ...         [1.000, 0.00],
            ...         [1.240, 2.30],
            ...         [1.585, 4.16],
            ...         [2.180, 6.00],
            ...         [3.020, 8.80],
            ...         [4.030, 12.5],
            ...         [4.760, 16.2],
            ...         [5.750, 23.6],
            ...         [6.850, 38.5],
            ...         [7.250, 49.6],
            ...         [7.600, 64.4],
            ...     ]
            ... ).T * np.array([[1.0], [0.0980665]])
            >>>
            >>> umat = fem.Hyperelastic(fem.anssari_benam_bucchi)
            >>> umat_new, res = umat.optimize(
            ...     ux=[λ, P], incompressible=True, relative=True
            ... )
            >>>
            >>> ux = np.linspace(λ.min(), λ.max(), num=50)
            >>> ax = umat_new.plot(incompressible=True, ux=ux, bx=None, ps=None)
            >>> ax.plot(λ, P, "C0x")

        See Also
        --------
        scipy.optimize.least_squares : Solve a nonlinear least-squares problem with
            bounds on the variables.

        References
        ----------
        .. [1] L. R. G. Treloar, "Stress-strain data for vulcanised rubber under various
           types of deformation", Transactions of the Faraday Society, vol. 40. Royal
           Society of Chemistry (RSC), p. 59, 1944. doi:
           `10.1039/tf9444000059 <https://doi.org/10.1039/tf9444000059>`_. Data
           available at https://www.uni-due.de/mathematik/ag_neff/neff_hencky.
        """
        from scipy.optimize import least_squares

        experiments = []
        for lc in [ux, bx, ps]:
            experiment = (None, None)
            if lc is not None:
                experiment = np.asarray(lc).reshape(2, -1)
            experiments.append(experiment)

        # get sizes of material parameters and offsets as cumulative sum
        offsets = np.cumsum([np.asarray(y).size for y in self.kwargs.values()])[:-1]

        # copy the material model formulation
        umat = self.copy()

        def fun(x):
            "Return the vector of residuals for given material parameters x."

            # update the material parameters
            for key, value in zip(umat.kwargs.keys(), np.split(x, offsets)):
                umat.kwargs[key] = value[0] if value.size == 1 else value

            # evaluate the load cases by the material model formulation
            model = umat.view(
                incompressible=incompressible,
                ux=experiments[0][0],
                bx=experiments[1][0],
                ps=experiments[2][0],
            ).evaluate()

            # calculate a list of residuals for each loadcase
            residuals = []
            for predicted, observed in zip(model, experiments):
                if observed[1] is not None:
                    res = predicted[1] - observed[1]
                    if relative:
                        observed_reference = np.array(observed[1])
                        observed_reference[observed_reference == 0] = 1
                        res /= observed_reference
                    residuals.append(res)

            return np.concatenate(residuals)

        # optimize the initial material parameters
        x0 = [np.asarray(value).ravel() for value in umat.kwargs.values()]
        res = least_squares(fun=fun, x0=np.concatenate(x0), **kwargs)

        def std(hessian, residuals_variance):
            "Return the estimated errors (standard deviations) of parameters."
            return np.sqrt(np.diag(np.linalg.inv(hessian) * residuals_variance))

        # estimate the optimization errors for each material parameter
        hess = res.jac.T @ res.jac
        res.dx = std(hess, 2 * res.cost / (len(res.fun) - len(res.x)))

        # save the final the material parameters
        for key, value in zip(umat.kwargs.keys(), np.split(res.x, offsets)):
            umat.kwargs[key] = value[0] if value.size == 1 else value

        return umat, res

    def is_stable(self, x, hessian=None):
        """Return a boolean mask for stability of isotropic material model formulations.

        At a given deformation gradient, a normal force is applied on each principal
        stretch direction. If the resulting incremental stretches are positive, the
        material model formulation is considered to be stable at the given deformation
        gradient.

        Parameters
        ----------
        x : list of ndarray
            The list with input arguments. These contain the extracted fields of a
            :class:`~felupe.FieldContainer`.
        hessian : ndarray or None, optional
            Second partial derivative of the strain energy density function w.r.t. the
            deformation gradient. Default is None.

        Returns
        -------
        ndarray
            Boolean mask of stability.

        Notes
        -----

        ..  warning::

            This stability check will lead to a singular matrix for isotropic
            (hyperelastic) material model formulations without a volumetric part.

        Examples
        --------
        First, let's check the stability of the Neo-Hookean material model formulation.
        The stability is evaluated on (valid) principal stretches of a biaxial
        deformation. All deformations are stable.

        ..  plot::

            >>> import numpy as np
            >>> import felupe as fem
            >>>
            >>> umat = fem.NeoHooke(mu=1.0, bulk=2.0)
            >>> view = umat.view()
            >>> λ = view.biaxial()[0]
            >>>
            >>> F = np.zeros((3, 3, 1, λ[0].size))
            >>> for a in range(3):
            ...     F[a, a] = λ[a]
            >>>
            >>> umat.is_stable([F])
            array([[ True,  True,  True,  True,  True,  True,  True,  True,  True,
                     True,  True,  True,  True,  True,  True,  True]])

        Now, let's check the stability of the Mooney-Rivlin material model formulation.
        The stability is evaluated on (valid) principal stretches of a biaxial
        deformation. Biaxial deformations are only stable up to a longitudinal stretch
        of 1.35.

        ..  plot::

            >>> import numpy as np
            >>> import felupe as fem
            >>> import felupe.constitution.tensortrax as mat
            >>>
            >>> umat = fem.Hyperelastic(
            ...     mat.models.hyperelastic.mooney_rivlin,
            ...     C10=0.25,
            ...     C01=0.25,
            ... ) & fem.Volumetric(bulk=5000)
            >>> view = umat.view()
            >>> λ = view.biaxial()[0]
            >>>
            >>> F = np.zeros((3, 3, 1, λ[0].size))
            >>> for a in range(3):
            ...     F[a, a] = λ[a]
            >>>
            >>> umat.is_stable([F])
            array([[ True,  True,  True,  True,  True,  True,  True,  True, False,
                    False, False, False, False, False, False, False]])

        """
        N = eigh(dot(transpose(x[0]), x[0]))[1]
        n = eigh(dot(x[0], transpose(x[0])))[1]

        if hessian is None:
            hessian = self.hessian(x)[0]

        A = np.einsum(
            "iJkL...,ia...,Ja...,kb...,Lb...->ab...",
            hessian,
            n,
            N,
            n,
            N,
            optimize=True,
        )

        stretch_change = np.diagonal(np.linalg.inv(A.T).T, axis1=0, axis2=1)

        return np.all(stretch_change > 0, axis=-1)

    def __and__(self, other_material):
        return CompositeMaterial(self, other_material)


def constitutive_material(Material, name=None):
    """A class-decorator for a constitutive material definition.

    Parameters
    ----------
    Material : object
        A class with methods for the gradient and the hessian of the strain energy
        density function per unit undeformed volume w.r.t. the deformation gradient
        tensor.
    name : str or None, optional
        The name of the derived class object. If None, the name is taken from
        ``Material`` (default is None).

    Returns
    -------
    object
        A derived class with multiple inheritance of
        :class:`~felupe.ConstitutiveMaterial` and ``Material``.

    Examples
    --------
    This example shows how to create a derived user material class to enable the
    methods from :class:`~felupe.ConstitutiveMaterial` on any (external) material.

    ..  plot::

        >>> import felupe as fem
        >>> import numpy as np
        >>>
        >>> class MyMaterialFormulation:
        ...     def __init__(self, a=5):
        ...         self.x = [np.zeros((3, 3))]
        ...         self.kwargs = {"a": a}
        ...
        ...     def gradient(self, x):
        ...         F, statevars = x[0], x[-1]
        ...         dWdF = self.kwargs["a"] * np.eye(3).reshape(3, 3, 1, 1)
        ...         return [dWdF, statevars]
        ...
        ...     def hessian(self, x, **kwargs):
        ...         F, statevars = x[0], x[-1]
        ...         d2WdFdF = self.kwargs["a"] * np.zeros((3, 3, 3, 3, 1, 1))
        ...         return [d2WdFdF]
        >>>
        >>> MyMaterial = fem.constitutive_material(MyMaterialFormulation)
        >>> umat = MyMaterial(a=0.5)
        >>> ax = umat.plot(incompressible=True)

    See Also
    --------
    felupe.ConstitutiveMaterial : Base class for constitutive material formulations.
    """
    if name is None:
        name = Material.__name__

    class ConstitutiveMaterialDerived(ConstitutiveMaterial, Material):
        pass

    ConstitutiveMaterialDerived.__name__ = name

    return ConstitutiveMaterialDerived


class CompositeMaterial(ConstitutiveMaterial):
    """A composite material with two constitutive materials merged. State variables are
    only considered for the first material.

    Parameters
    ----------
    material : ConstitutiveMaterial
        First constitutive material.
    other_material : ConstitutiveMaterial
        Second constitutive material.

    Notes
    -----
    ..  warning::
        Do not merge two constitutive materials with the same keys of material
        parameters. In this case, the values of these material parameters are taken from
        the first constitutive material.

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> nh = fem.NeoHooke(mu=1.0)
        >>> vol = fem.Volumetric(bulk=2.0)
        >>> umat = nh & vol
        >>> ax = umat.plot()

    """

    def __init__(self, material, other_material):
        self.materials = [material, other_material]
        self.kwargs = {**other_material.kwargs, **material.kwargs}
        self.x = material.x

    def gradient(self, x, **kwargs):
        gradients = [material.gradient(x, **kwargs) for material in self.materials]
        nfields = max(1, len(x) - 1)
        P = [np.sum([grad[i] for grad in gradients], axis=0) for i in range(nfields)]
        statevars_new = gradients[0][-1]
        return [*P, statevars_new]

    def hessian(self, x, **kwargs):
        hessians = [material.hessian(x, **kwargs) for material in self.materials]
        nfields = max(1, len(x) - 1)
        return [np.sum([hess[i] for hess in hessians], axis=0) for i in range(nfields)]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import array, diag, log, sqrt, trace
from jax.numpy.linalg import det, eigvalsh

from ....tensortrax.models.hyperelastic import extended_tube as extended_tube_docstring


@wraps(extended_tube_docstring)
def extended_tube(C, Gc, delta, Ge, beta):
    J3 = det(C) ** (-1 / 3)
    D = J3 * trace(C)
    λ1, λ2, λ3 = sqrt(J3 * eigvalsh(C + diag(array([0, 1e-4, -1e-4]))))
    β = beta
    δ = delta
    γ = (1 - δ**2) * (D - 3) / (1 - δ**2 * (D - 3))
    Wc = Gc / 2 * (γ + log(1 - δ**2 * (D - 3)))
    We = 2 * Ge / β**2 * (λ1**-β + λ2**-β + λ3**-β - 3)
    return Wc + We



from ._chain import langevin, linear
from ._framework_affine import (
    affine_stretch,
    affine_stretch_statevars,
    affine_tube,
    affine_tube_statevars,
)
from ._framework_nonaffine import nonaffine_stretch, nonaffine_tube

__all__ = [
    "affine_stretch",
    "affine_stretch_statevars",
    "affine_tube",
    "affine_tube_statevars",
    "linear",
    "langevin",
    "nonaffine_stretch",
    "nonaffine_tube",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ..element import lagrange_hexahedron, lagrange_quad
from ._line_rectangle_cube import cube_hexa, line_line, rectangle_quad
from ._mesh import Mesh
from ._tools import concatenate


class Point(Mesh):
    """A mesh with a single vertex point located at ``a``.

    Parameters
    ----------
    a : float, optional
        Vertex point coordinate of the mesh (default is 0.0).

    Examples
    --------
    ..  pyvista-plot::

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Point(a=-2.1)
        >>> mesh
        <felupe Mesh object>
          Number of points: 1
          Number of cells:
            vertex: 1

        >>> mesh.points
        array([[-2.1]])

        >>> mesh.cells
        array([[0]])
    """

    def __init__(self, a=0.0):
        self.a = a

        points = np.array([a]).reshape(1, -1)
        cells = np.array([[0]])
        cell_type = "vertex"

        super().__init__(points, cells, cell_type)


class Line(Mesh):
    """A 1d-mesh with lines between ``a`` and ``b`` with ``n`` points.

    Parameters
    ----------
    a : float, optional
        Left end point of the mesh (default is 0.0).
    b : float, optional
        Right end point of the mesh (default is 1.0).
    n : int, optional
        Number of points (default is 2).

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.mesh.Line(a=-2.1, b=3.5, n=3)
       >>> mesh.plot(line_width=8).show()

    >>> mesh
    <felupe Mesh object>
      Number of points: 3
      Number of cells:
        line: 2

    >>> mesh.points
    array([[-2.1],
           [ 0.7],
           [ 3.5]])

    >>> mesh.cells
    array([[0, 1],
           [1, 2]])
    """

    def __init__(self, a=0.0, b=1.0, n=2):
        self.a = a
        self.b = b
        self.n = n

        points, cells, cell_type = line_line(a, b, n)

        super().__init__(points, cells, cell_type)


class Rectangle(Mesh):
    """A rectangular 2d-mesh with quads between ``a`` and ``b`` with ``n``
    points per axis.

    Parameters
    ----------
    a : 2-tuple of float, optional
        Lower-left end point of the mesh (default is (0.0, 0.0)).
    b : 2-tuple of float, optional
        Upper-right end point of the mesh (default is (1.0, 1.0)).
    n : int or 2-tuple of int, optional
        Number of points per axis (default is (2, 2)).

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.mesh.Rectangle(a=(-1.2, 0.5), b=(4.5, 7.3), n=3)
       >>> mesh.plot().show()

    >>> mesh
    <felupe Mesh object>
      Number of points: 9
      Number of cells:
        quad: 4

    >>> mesh.points
    array([[-1.2 ,  0.5 ],
           [ 1.65,  0.5 ],
           [ 4.5 ,  0.5 ],
           [-1.2 ,  3.9 ],
           [ 1.65,  3.9 ],
           [ 4.5 ,  3.9 ],
           [-1.2 ,  7.3 ],
           [ 1.65,  7.3 ],
           [ 4.5 ,  7.3 ]])

    >>> mesh.cells
    array([[0, 1, 4, 3],
           [1, 2, 5, 4],
           [3, 4, 7, 6],
           [4, 5, 8, 7]])
    """

    def __init__(self, a=(0.0, 0.0), b=(1.0, 1.0), n=(2, 2)):
        self.a = a
        self.b = b
        self.n = n

        points, cells, cell_type = rectangle_quad(a, b, n)

        super().__init__(points, cells, cell_type)


class Cube(Mesh):
    """A cube shaped 3d-mesh with hexahedrons between ``a`` and ``b`` with ``n``
    points per axis.

    Parameters
    ----------
    a : 3-tuple of float, optional
        Lower-left end point of the mesh (default is (0.0, 0.0, 0.0)).
    b : 3-tuple of float, optional
        Upper-right end point of the mesh (default is (1.0, 1.0, 1.0)).
    n : int or 3-tuple of int, optional
        Number of points per axis (default is (2, 2, 2)).

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.mesh.Cube(a=(-1.2, 0.5, 6.2), b=(4.5, 7.3, 9.3), n=(3, 2, 2))
       >>> mesh.plot().show()

    >>> mesh
    <felupe Mesh object>
      Number of points: 12
      Number of cells:
        hexahedron: 2

    >>> mesh.points
    array([[-1.2 ,  0.5 ,  6.2 ],
           [ 1.65,  0.5 ,  6.2 ],
           [ 4.5 ,  0.5 ,  6.2 ],
           [-1.2 ,  7.3 ,  6.2 ],
           [ 1.65,  7.3 ,  6.2 ],
           [ 4.5 ,  7.3 ,  6.2 ],
           [-1.2 ,  0.5 ,  9.3 ],
           [ 1.65,  0.5 ,  9.3 ],
           [ 4.5 ,  0.5 ,  9.3 ],
           [-1.2 ,  7.3 ,  9.3 ],
           [ 1.65,  7.3 ,  9.3 ],
           [ 4.5 ,  7.3 ,  9.3 ]])

    >>> mesh.cells
    array([[ 0,  1,  4,  3,  6,  7, 10,  9],
           [ 1,  2,  5,  4,  7,  8, 11, 10]])
    """

    def __init__(self, a=(0.0, 0.0, 0.0), b=(1.0, 1.0, 1.0), n=(2, 2, 2)):
        self.a = a
        self.b = b
        self.n = n

        points, cells, cell_type = cube_hexa(a, b, n)

        super().__init__(points, cells, cell_type)


class Grid(Mesh):
    """A grid shaped 3d-mesh with hexahedrons. Basically a wrapper for
    :func:`numpy.meshgrid` with  default ``indexing="ij"``.

    Parameters
    ----------
    x1, x2,..., xn : array_like
        1-D arrays representing the coordinates of a grid.

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import numpy as np
       >>> import felupe as fem
       >>>
       >>> x1 = np.linspace(0, 2, 3)**2
       >>> x2 = np.sqrt(np.linspace(0, 1, 3))
       >>>
       >>> mesh = fem.mesh.Grid(x1, x2)
       >>> mesh.plot().show()

    >>> mesh
    <felupe Mesh object>
      Number of points: 9
      Number of cells:
        quad: 4

    >>> mesh.points
    array([[0.        , 0.        ],
           [1.        , 0.        ],
           [4.        , 0.        ],
           [0.        , 0.70710678],
           [1.        , 0.70710678],
           [4.        , 0.70710678],
           [0.        , 1.        ],
           [1.        , 1.        ],
           [4.        , 1.        ]])

    >>> mesh.cells
    array([[0, 1, 4, 3],
           [1, 2, 5, 4],
           [3, 4, 7, 6],
           [4, 5, 8, 7]])

    See Also
    --------
    numpy.meshgrid : Return a list of coordinate matrices from coordinate vectors.
    """

    def __init__(self, *xi, indexing="ij", **kwargs):
        shape = np.array([len(x) for x in xi])
        n = shape if len(shape) > 1 else shape[0]

        M = [None, line_line, rectangle_quad, cube_hexa][len(xi)]
        points, cells, cell_type = M(n=n)

        super().__init__(points, cells, cell_type)

        self.points = (
            np.vstack(
                [g.T.ravel() for g in np.meshgrid(*xi, indexing=indexing, **kwargs)]
            )
            .astype(float)
            .T
        )


class RectangleArbitraryOrderQuad(Rectangle):
    """A rectangular 2d-mesh with an arbitrary-order Lagrange quad between ``a`` and
    ``b``.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.mesh.RectangleArbitraryOrderQuad(order=5).add_runouts()
        >>> mesh.plot(
        >>>     nonlinear_subdivision=4,
        >>>     plotter=mesh.plot(style="points", color="black"),
        >>> ).show()
    """

    def __init__(self, a=(0.0, 0.0), b=(1.0, 1.0), order=2):
        super().__init__(a=a, b=b, n=order + 1)
        self.update(
            cells=lagrange_quad(order=order).reshape(1, -1),
            cell_type="VTK_LAGRANGE_QUADRILATERAL",
        )


class CubeArbitraryOrderHexahedron(Cube):
    """A rectangular 3d-mesh with an arbitrary-order Lagrange hexahedron between ``a``
    and ``b``.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.mesh.CubeArbitraryOrderHexahedron(order=5).add_runouts()
        >>> mesh.plot(
        >>>     nonlinear_subdivision=4,
        >>>     plotter=mesh.plot(style="points", color="black"),
        >>> ).show()
    """

    def __init__(self, a=(0.0, 0.0, 0.0), b=(1.0, 1.0, 1.0), order=2):
        super().__init__(a=a, b=b, n=order + 1)
        self.update(
            cells=lagrange_hexahedron(order=order).reshape(1, -1),
            cell_type="VTK_LAGRANGE_HEXAHEDRON",
        )


class Circle(Mesh):
    """A circular shaped 2d-mesh with quads and ``n`` points on the circumferential
    edge of a 45-degree section. 90-degree ``sections`` are placed at given angles in
    degree.

    Parameters
    ----------
    radius : float, optional
        Lower-left end point of the mesh (default is (0.0, 0.0)).
    centerpoint : 2-list of float, optional
        Coordinates of the origin where the circle is centered (default is [1.0, 1.0]).
    n : int, optional
        Number of points per axis for a quarter of the embedded rectangle
        (default is 2).
    sections : list of int or float, optional
        Rotation angles in deg where quarter circles (sections) are placed (default is
        [0, 90, 180, 270]).
    value : float
        First shape parameter of the embedded rectangle (default is 0.15).
    exponent : int
        Second shape parameter of the embedded rectangle (default is 2).
    decimals : int
        Decimals used for rounding point coordinates to avoid non-connected sections
        (default is 10).

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.mesh.Circle()
       >>> mesh.plot().show()

    >>> mesh
    <felupe Mesh object>
      Number of points: 17
      Number of cells:
        quad: 12

    >>> mesh.points
    array([[-1.        ,  0.        ],
           [-0.70710678, -0.70710678],
           [-0.70710678,  0.70710678],
           [-0.5       ,  0.        ],
           [-0.425     , -0.425     ],
           [-0.425     ,  0.425     ],
           [ 0.        , -1.        ],
           [ 0.        , -0.5       ],
           [ 0.        ,  0.        ],
           [ 0.        ,  0.5       ],
           [ 0.        ,  1.        ],
           [ 0.425     , -0.425     ],
           [ 0.425     ,  0.425     ],
           [ 0.5       ,  0.        ],
           [ 0.70710678, -0.70710678],
           [ 0.70710678,  0.70710678],
           [ 1.        ,  0.        ]])

    >>> mesh.cells
    array([[12, 13, 16, 15],
           [15, 10,  9, 12],
           [ 8, 13, 12,  9],
           [ 5,  9, 10,  2],
           [ 2,  0,  3,  5],
           [ 8,  9,  5,  3],
           [ 4,  3,  0,  1],
           [ 1,  6,  7,  4],
           [ 8,  3,  4,  7],
           [11,  7,  6, 14],
           [14, 16, 13, 11],
           [ 8,  7, 11, 13]])
    """

    def __init__(
        self,
        radius=1.0,
        centerpoint=[0.0, 0.0],
        n=2,
        sections=[0, 90, 180, 270],
        value=0.15,
        exponent=2,
        decimals=10,
    ):
        rect = Rectangle(b=(0.5, 0.5), n=(n, n))
        right = rect.points[:, 0] == 0.5

        rect = rect.add_runouts(values=[-value], axis=0, exponent=exponent)
        rect.points[:, 0] = rect.points[:, 1].reshape(n, n, order="F").ravel()

        line = Line(n=n)
        phi = np.linspace(np.pi / 4, 0, n)

        bottom = line.copy(points=rect.points[right][::-1])
        top = line.copy(points=np.vstack([np.cos(phi), np.sin(phi)]).T)
        face = bottom.fill_between(top, n=n)

        quarter = concatenate([face, face.mirror(normal=[-1, 1]), rect])
        circle = concatenate([quarter.rotate(alpha, 2) for alpha in sections]).sweep(
            decimals=decimals
        )

        circle.points *= radius
        circle.points += np.array(centerpoint)

        super().__init__(
            points=circle.points, cells=circle.cells, cell_type=circle.cell_type
        )


class Triangle(Mesh):
    """A triangular shaped 2d-mesh with quads and ``n`` points at the edges of the three
    sub-quadrilaterals.

    Parameters
    ----------
    a : 2-tuple of float, optional
        First end point of the mesh (default is (0.0, 0.0)).
    b : 2-tuple of float, optional
        Second end point of the mesh (default is (1.0, 0.0)).
    c : 2-tuple of float, optional
        Third end point of the mesh (default is (0.0, 1.0)).
    n : int, optional
        Number of points per axis is 2n - 1 if n >= 2 (default is 2).
    decimals : int
        Decimals used for rounding point coordinates to avoid non-connected sections
        (default is 10).

    Notes
    -----
    If ``n < 2``, a triangle mesh with a single triangle cell is created.

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.mesh.Triangle(a=(-0.1, 0.2), b=(1.2, 0.1), c=(0.1, 0.9), n=1)
       >>> mesh.plot().show()

    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.mesh.Triangle(a=(-0.1, 0.2), b=(1.2, 0.1), c=(0.1, 0.9), n=2)
       >>> mesh.plot().show()

    >>> mesh
    <felupe Mesh object>
      Number of points: 19
      Number of cells:
        quad: 12

    >>> mesh.points
    array([[0.1       , 0.9       ],
           [0.15      , 0.725     ],
           [0.2       , 0.55      ],
           [0.25      , 0.375     ],
           [0.3       , 0.2       ],
           [0.36666667, 0.475     ],
           [0.37083333, 0.5875    ],
           [0.375     , 0.7       ],
           [0.44583333, 0.325     ],
           [0.525     , 0.175     ],
           [0.53333333, 0.4       ],
           [0.59166667, 0.45      ],
           [0.64166667, 0.275     ],
           [0.65      , 0.5       ],
           [0.75      , 0.15      ],
           [0.78333333, 0.2875    ],
           [0.925     , 0.3       ],
           [0.975     , 0.125     ],
           [1.2       , 0.1       ]])

    >>> mesh.cells
    array([[14, 12,  8,  9],
           [12, 10,  5,  8],
           [ 9,  8,  3,  4],
           [ 8,  5,  2,  3],
           [18, 16, 15, 17],
           [16, 13, 11, 15],
           [17, 15, 12, 14],
           [15, 11, 10, 12],
           [10, 11,  6,  5],
           [11, 13,  7,  6],
           [ 5,  6,  1,  2],
           [ 6,  7,  0,  1]])
    """

    def __init__(
        self,
        a=(0.0, 0.0),
        b=(1.0, 0.0),
        c=(0.0, 1.0),
        n=2,
        decimals=10,
    ):
        if n < 2:  # triangle mesh
            points = np.vstack([a, b, c])
            cells = np.arange(3).reshape(1, 3)
            cell_type = "triangle"

        else:  # quad mesh
            a = np.asarray(a)
            b = np.asarray(b)
            c = np.asarray(c)

            sections = []

            centerpoint = (a + b + c) / 3
            centerpoints = {"ab": (a + b) / 2, "bc": (b + c) / 2, "ac": (a + c) / 2}

            line = Line(n=n)

            # section (connected to point) a
            x1 = np.linspace(a[0], centerpoints["ac"][0], n)
            y1 = np.linspace(a[1], centerpoints["ac"][1], n)

            left = line.copy(points=np.vstack([x1, y1]).T)

            x2 = np.linspace(centerpoints["ab"][0], centerpoint[0], n)
            y2 = np.linspace(centerpoints["ab"][1], centerpoint[1], n)

            middle = line.copy(points=np.vstack([x2, y2]).T)

            sections.append(middle.fill_between(left, n=n))

            # section (connected to point) b
            x3 = np.linspace(b[0], centerpoints["bc"][0], n)
            y3 = np.linspace(b[1], centerpoints["bc"][1], n)

            right = line.copy(points=np.vstack([x3, y3]).T)

            sections.append(right.fill_between(middle, n=n))

            # section (connected to point) c
            x4 = np.linspace(centerpoints["ac"][0], c[0], n)
            y4 = np.linspace(centerpoints["ac"][1], c[1], n)

            top = line.copy(points=np.vstack([x4, y4]).T)

            x5 = np.linspace(centerpoint[0], centerpoints["bc"][0], n)
            y5 = np.linspace(centerpoint[1], centerpoints["bc"][1], n)

            bottom = line.copy(points=np.vstack([x5, y5]).T)

            sections.append(bottom.fill_between(top, n=n))

            # combine sections
            triangle = concatenate(sections).sweep(decimals=decimals)

            points = triangle.points
            cells = triangle.cells
            cell_type = triangle.cell_type

        super().__init__(points=points, cells=cells, cell_type=cell_type)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from tensortrax import Tensor, Δ, Δδ, f, δ
from tensortrax.math import exp, log, trace
from tensortrax.math.linalg import det


def alexander(C, C1, C2, C3, gamma, k):
    r"""Strain energy function of the isotropic hyperelastic
    `Alexander <https://doi.org/10.1016/0020-7225(68)90006-2>`_ material
    formulation [1]_.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    C1 : float
        Scale factor for the first invariant term.
    C2 : float
        Scale factor for the second invariant term.
    C3 : float
        Scale factor for the logarithmic second invariant term.
    gamma : float
        Offset-normalization parameter for the logarithmic second invariant term.
    k : float
        Scale factor for the exponential first invariant term.

    Notes
    -----
    ..  warning::
        The strain energy function of the Alexander model formulation is not directly
        implemented. Only its gradient and hessian w.r.t. the right Cauchy-Green
        deformation tensor are defined. This is because the imaginary error-function
        :math:`\text{erfi}(x)` is not included in NumPy - this would require SciPy as a
        dependency.

    The strain energy function is given in Eq. :eq:`psi-alexander`

    ..  math::
        :label: psi-alexander

        \psi = C_1 \int_{\hat{I}_1} \exp \left(
                   k \left(\hat{I}_1 - 3 \right)^2
               \right) \ d\hat{I}_1
             + C_2 \ln \left(\frac{\hat{I}_2 - 3 + \gamma}{\gamma} \right)
             + C_3 \left(\hat{I}_2 - 3 \right)

    with the first and second main invariant of the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`invariants-alexander`.

    ..  math::
        :label: invariants-alexander

        \hat{I}_1 &= J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

        \hat{I}_2 &= J^{-4/3} \frac{1}{2} \left(
            \text{tr}\left(\boldsymbol{C}\right)^2 -
            \text{tr}\left(\boldsymbol{C}^2\right)
        \right)

    The initial shear modulus :math:`\mu` is given in Eq.
    :eq:`shear-modulus-alexander`.

    ..  math::
        :label: shear-modulus-alexander

        \mu = 2 \left( C_1 + \frac{C_2}{\gamma} + C_3 \right)

    Examples
    --------

    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(
        ...    fem.alexander, C1=17, C2=19.85, C3=1, gamma=0.735, k=0.00015
        ... )
        >>> ux = fem.math.linsteps([0.6, 5], num=50)
        >>> ps = fem.math.linsteps([1, 5], num=50)
        >>> bx = fem.math.linsteps([1, 3], num=50)
        >>>
        >>> ax = umat.plot(ux=ux, ps=ps, bx=bx, incompressible=True)

    References
    ----------
    .. [1] H. Alexander, "A constitutive relation for rubber-like materials",
       International Journal of Engineering Science, vol. 6, no. 9. Elsevier BV, pp.
       549–563, Sep. 1968. doi:
       `10.1016/0020-7225(68)90006-2 <https://doi.org/10.1016/0020-7225(68)90006-2>`_.
    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (I1**2 - J3**2 * trace(C @ C)) / 2

    def W(I1):
        "A non-evaluated function with defined first- and second derivatives."

        dWdI1 = exp(k * (I1 - 3) ** 2)
        return Tensor(
            x=np.full_like(f(I1), fill_value=np.nan),
            δx=f(dWdI1) * δ(I1),
            Δx=f(dWdI1) * Δ(I1),
            Δδx=δ(dWdI1) * Δ(I1) + f(dWdI1) * Δδ(I1),
            ntrax=I1.ntrax,
        )

    return C1 * W(I1) + C2 * log(((I2 - 3) + gamma) / gamma) + C3 * (I2 - 3)
